from insuree.models import FamilyType, ConfirmationType


class GroupTypeMapping(object):

    group_type = {}
    @classmethod
    def load(cls):
        cls.group_type = {
            str(family_type.code): family_type.type for family_type in FamilyType.objects.all()
        }


class ConfirmationTypeMapping(object):

    confirmation_type = {}
    @classmethod
    def load(cls):
        cls.confirmation_type = {
            str(confirmation_type.code): confirmation_type.confirmationtype for confirmation_type in ConfirmationType.objects.all()
        }



from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.tests import GenericFhirAPITestMixin, FhirApiReadTestMixin
from rest_framework.test import APITestCase
from api_fhir_r4.tests.mixin.logInMixin import LogInMixin
from graphql_jwt.shortcuts import get_token
from core.test_helpers import create_test_interactive_user
from dataclasses import dataclass
from core.models import User
from rest_framework import status
from policyholder.tests.helpers import create_test_policy_holder

@dataclass
class DummyContext:
    """ Just because we need a context to generate. """
    user: User

class OrganisationAPITests(GenericFhirAPITestMixin, FhirApiReadTestMixin, APITestCase, LogInMixin):
    base_url = GeneralConfiguration.get_base_url() + 'Organization/'
    _test_json_path = None
    _test_request_data_credentials = None
    admin_token = None 
    admin_user = None

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.test_policy_holder = create_test_policy_holder()
        cls.admin_user = create_test_interactive_user(username="testLocationAdmin")
        cls.admin_token = get_token(cls.admin_user, DummyContext(user=cls.admin_user))


    def test_simple_list_page_2(self):
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {self.admin_token}"
        }
        response = self.client.get(self.base_url+ '?page-offset=2', format='json', **headers)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNotNone(response.content)
    
    def test_simple_ph(self):
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {self.admin_token}"
        }
        #
        response = self.client.get(self.base_url+ str(self.test_policy_holder.uuid).upper()+ '/' , format='json', **headers)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNotNone(response.content)        
        
    def get_or_create_user_api(self):
        user = DbManagerUtils.get_object_or_none(User, username=self._TEST_USER_NAME)
        if user is None:
            user = self.__create_user_interactive_core()
        return user



# TODO uncomment if someone need this converter with connection to openHIM

"""
from policy.services import ByInsureeRequest

from api_fhir_r4.configurations import R4CoverageEligibilityConfiguration as Config
from api_fhir_r4.converters import BaseFHIRConverter, PatientConverter
from fhir.resources.R4B.money import Money
from fhir.resources.R4B.coverageeligibilityresponse import CoverageEligibilityResponse as FHIREligibilityResponse, \
    CoverageEligibilityResponseInsuranceItem, CoverageEligibilityResponseInsurance, CoverageEligibilityResponseInsuranceItemBenefit
from api_fhir_r4.models import CoverageEligibilityRequestV2 as FHIREligibilityRequest
from api_fhir_r4.utils import TimeUtils
"""
"""
class PolicyCoverageEligibilityRequestConverter(BaseFHIRConverter):

    @classmethod
    def to_fhir_obj(cls, eligibility_response):
        fhir_response = FHIREligibilityResponse.construct()
        try:
            for item in eligibility_response.items:
                if item.status in Config.get_fhir_active_policy_status():
                    cls.build_fhir_insurance(fhir_response, item)
        except:
            for item in eligibility_response['items']:
                if type(fhir_response.insurance) is not list:
                    fhir_response.insurance = [item]
                else:
                    fhir_response.insurance.append(item)
        return fhir_response

    @classmethod
    def to_imis_obj(cls, fhir_eligibility_request, audit_user_id):
        fhir_eligibility_request["status"] = "active"
        fhir_eligibility_request["purpose"] = ["validation"]
        fhir_eligibility_request["created"] = TimeUtils.date().isoformat()
        fhir_eligibility_request = FHIREligibilityRequest(**fhir_eligibility_request)
        uuid = cls.build_imis_uuid(fhir_eligibility_request)
        return ByInsureeRequest(uuid)

    @classmethod
    def build_fhir_insurance(cls, fhir_response, response):
        result = CoverageEligibilityResponseInsurance.construct()
        #cls.build_fhir_insurance_contract(result, response)
        cls.build_fhir_money_item(result, Config.get_fhir_balance_code(),
                                  response.ceiling,
                                  response.ded)
        if type(fhir_response.insurance) is not list:
            fhir_response.insurance = [result]
        else:
            fhir_response.insurance.append(result)

    '''
    @classmethod
    def build_fhir_insurance_contract(cls, insurance, contract):
        insurance.contract = ContractConverter.build_fhir_resource_reference(
            contract)
    '''

    @classmethod
    def build_fhir_money_item(cls, insurance, code, allowed_value, used_value):
        item = cls.build_fhir_generic_item(code)
        cls.build_fhir_money_item_benefit(
            item, allowed_value, used_value)
        if type(insurance.item) is not list:
            insurance.item = [item]
        else:
            insurance.item.append(item)

    @classmethod
    def build_fhir_generic_item(cls, code):
        item = CoverageEligibilityResponseInsuranceItem.construct()
        item.category = cls.build_simple_codeable_concept(
            Config.get_fhir_balance_default_category())
        return item

    @classmethod
    def build_fhir_money_item_benefit(cls, item, allowed_value, used_value):
        benefit = cls.build_fhir_generic_item_benefit()
        allowed_money_value = Money.construct()
        allowed_money_value.value = allowed_value or 0
        benefit.allowedMoney = allowed_money_value
        used_money_value = Money.construct()
        used_money_value.value = used_value or 0
        benefit.usedMoney = used_money_value
        if type(item.benefit) is not list:
            item.benefit = [benefit]
        else:
            item.benefit.append(benefit)

    @classmethod
    def build_fhir_generic_item_benefit(cls):
        benefit = CoverageEligibilityResponseInsuranceItemBenefit()
        benefit.type = cls.build_simple_codeable_concept(
            Config.get_fhir_financial_code())
        return benefit

    @classmethod
    def build_imis_uuid(cls, fhir_eligibility_request):
        uuid = None
        patient_reference = fhir_eligibility_request.patient
        if patient_reference:
            uuid = PatientConverter.get_resource_id_from_reference(
                patient_reference)
        return uuid
"""



from rest_framework.serializers import Serializer


class MultiSerializerSerializerClass(Serializer):
    """
     Serves as base serializer class for instances using multiserializer mixin.
     """

    def update(self, instance, validated_data):
        raise NotImplementedError("MultiSerializerSerializerClass `update` not supported. Should be"
                                  "performed by contained serializers")

    def create(self, validated_data):
        raise NotImplementedError("MultiSerializerSerializerClass `create` not supported. Should be"
                                  "performed by contained serializers")

    def to_representation(self, instance):
        """
        This override is required by rest framework serializer logic. During render the serilaizer to_representation()
        is called. However multiserializer aggregate output data using mixins. Therefore by default IMIS FHIR approach
        calls BaseFhirSerializer to_representation() which cause not implemented exception.
        :param instance:
        :return:
        """
        return {}



from django.utils.translation import gettext as _
from rest_framework.test import APITestCase

from api_fhir_r4.configurations import R4IdentifierConfig
from fhir.resources.R4B.medication import Medication as FHIRMedication
from api_fhir_r4.converters import MedicationConverter
from api_fhir_r4.tests import GenericFhirAPITestMixin, \
    FhirApiCreateTestMixin, FhirApiUpdateTestMixin, FhirApiReadTestMixin
from api_fhir_r4.configurations import GeneralConfiguration


class MedicationAPITests(GenericFhirAPITestMixin, FhirApiCreateTestMixin, FhirApiUpdateTestMixin, FhirApiReadTestMixin, APITestCase):

    base_url = GeneralConfiguration.get_base_url()+'Medication/'
    _test_json_path = "/test/test_medication.json"
    _TEST_EXPECTED_CODE = "TESTT"

    def setUp(self):
        super(MedicationAPITests, self).setUp()

    def verify_updated_obj(self, updated_obj):
        self.assertTrue(isinstance(updated_obj, FHIRMedication))
        code = MedicationConverter.get_fhir_identifier_by_code(
            updated_obj.identifier,
            R4IdentifierConfig.get_fhir_generic_type_code()
        )
        self.assertEqual(self._TEST_EXPECTED_CODE, code)

    def update_resource(self, data):
        data['identifier'][0]["value"] = self._TEST_EXPECTED_CODE

    def update_payload_missing_code_identifier(self, data):
        for i in range(len(data["identifier"])):
            if data["identifier"][i]["type"]["coding"][0]["code"] == "Code":
                del data["identifier"][i]
                return data

    def update_payload_missing_item_name(self, data):
        if "code" in data:
            data.pop("code")
        return data

    def test_post_should_raise_error_no_code_identifier(self):
        self.login()
        self.create_dependencies()
        modified_payload = self.update_payload_missing_code_identifier(data=self._test_request_data)
        response = self.client.post(self.base_url, data=modified_payload, format='json')
        response_json = response.json()
        splited_output = self.get_response_details(response_json).split(" ")
        self.assertEqual(
            self.get_response_details(response_json),
            _("The request cannot be processed due to the following issues:\nMissing medication `item_code` attribute")
        )

    def test_post_should_raise_error_no_item_name(self):
        self.login()
        self.create_dependencies()
        modified_payload = self.update_payload_missing_item_name(data=self._test_request_data)
        response = self.client.post(self.base_url, data=modified_payload, format='json')
        response_json = response.json()
        splited_output = self.get_response_details(response_json).split(" ")
        self.assertEqual(
            self.get_response_details(response_json),
            _("The request cannot be processed due to the following issues:\nMissing medication `item_name` attribute")
        )



from rest_framework import viewsets

from api_fhir_r4.mixins import (
    MultiIdentifierRetrieverMixin,
    MultiIdentifierUpdateMixin
)
from api_fhir_r4.model_retrievers import (
    UUIDIdentifierModelRetriever,
    GroupIdentifierModelRetriever
)
from api_fhir_r4.paymentNotice import PaymentNoticeSerializer
from api_fhir_r4.permissions import FHIRApiPaymentPermissions
from api_fhir_r4.views.fhir.base import BaseFHIRView
from api_fhir_r4.views.filters import DateUpdatedRequestParameterFilter
from invoice.models import PaymentInvoice


class PaymentNoticeViewSet(BaseFHIRView, MultiIdentifierRetrieverMixin,
                   MultiIdentifierUpdateMixin, viewsets.ModelViewSet):
    retrievers = [UUIDIdentifierModelRetriever, GroupIdentifierModelRetriever]
    serializer_class = PaymentNoticeSerializer
    permission_classes = (FHIRApiPaymentPermissions,)

    def list(self, request, *args, **kwargs):
        queryset = self.get_queryset()
        identifier = request.GET.get("identifier")
        if identifier:
            return self.retrieve(request, *args, **{**kwargs, 'identifier': identifier})
        else:
            queryset = queryset.filter(is_deleted=False)
        serializer = PaymentNoticeSerializer(self.paginate_queryset(queryset), many=True)
        return self.get_paginated_response(serializer.data)

    def retrieve(self, *args, **kwargs):
        response = super().retrieve(self, *args, **kwargs)
        return response

    def get_queryset(self):
        queryset = PaymentInvoice.objects.filter(is_deleted=False).order_by('date_created')
        return DateUpdatedRequestParameterFilter(self.request).filter_queryset(queryset)



from fhir.resources.R4B.practitionerrole import PractitionerRole
from fhir.resources.R4B.extension import Extension
from fhir.resources.R4B.identifier import Identifier
from fhir.resources.R4B.reference import Reference
from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig
from api_fhir_r4.converters import EnrolmentOfficerPractitionerConverter, EnrolmentOfficerPractitionerRoleConverter
from api_fhir_r4.tests import GenericTestMixin, EnrolmentOfficerPractitionerTestMixin, LocationTestMixin
from core.models import Officer
from location.test_helpers import create_test_village 
from core.test_helpers import create_test_officer

class EnrolmentOfficerPractitionerRoleTestMixin(GenericTestMixin):
    test_officer = None
    test_village = None
    test_substitution_officer = None
    _TEST_SUBSTITUTION_OFFICER_UUID = "f4bf924a-e2c9-46dc-9fa3-d54a1a67ea86"
    _TEST_OFFICER_UUID = "b578a621-0b11-4889-9454-a8e498c35dee"
    _TEST_SUBSTITUTION_OFFICER_CODE = "EOTESB"
    _TEST_OFFICER_CODE = "EOTEST"
    _TEST_PRACTITIONER_REFERENCE = None
    _TEST_VILLAGE_REFERENCE = 'Location/20fcfa89-149a-429e-bd26-b014b95fbb2d'
    _TEST_VILLAGE_UUID = "041890f3-d90b-46ca-8b25-56c3f8d60615"
    sub_str = {}
    
    def setUp(self):
        self.test_village = create_test_village()
        self.test_substitution_officer   = create_test_officer(custom_props ={
                "uuid": self._TEST_SUBSTITUTION_OFFICER_UUID,
                "code": self._TEST_SUBSTITUTION_OFFICER_CODE,
                "last_name": "Officer",
                "other_names": "Test",
                'location':self.test_village.parent.parent
        })
        self.test_officer= create_test_officer(custom_props ={                
                "uuid": self._TEST_OFFICER_UUID,
                "code": self._TEST_OFFICER_CODE,
                "substitution_officer": self.test_substitution_officer,
                "last_name": "Officer",
                "other_names": "Test",
                "validity_to": None,
                "audit_user_id": -1,
                'location':self.test_village.parent.parent})
        
        self._TEST_PRACTITIONER_REFERENCE = "Practitioner/" + str(self.test_officer.uuid)

        self.sub_str[self._TEST_SUBSTITUTION_OFFICER_UUID]=self.test_substitution_officer.uuid
        self.sub_str[self._TEST_OFFICER_UUID]=self.test_officer.uuid
        self.sub_str[self._TEST_VILLAGE_UUID]=self.test_village.uuid
        self.sub_str[self._TEST_VILLAGE_REFERENCE]="Location/" + str(self.test_village.uuid)
        self._TEST_VILLAGE_REFERENCE = "Location/" + str(self.test_village.uuid)




    def verify_imis_instance(self, imis_obj):
        self.assertEqual(self.test_officer.substitution_officer.code, imis_obj.substitution_officer.code)
        self.assertEqual(self.test_village.code, imis_obj.location.code)

    def create_test_imis_instance(self):
        return self.test_officer

    def create_test_fhir_instance(self):
        self.create_test_imis_instance()
        fhir_practitioner_role = PractitionerRole.construct()
        identifiers = []
        code = EnrolmentOfficerPractitionerRoleConverter.build_fhir_identifier(
            self.test_officer.code,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            R4IdentifierConfig.get_fhir_generic_type_code()
        )
        identifiers.append(code)
        location_reference = Reference.construct()
        location_reference.reference = self._TEST_VILLAGE_REFERENCE
        fhir_practitioner_role.location = [location_reference]
        practitioner_reference = Reference.construct()
        practitioner_reference.reference = self._TEST_PRACTITIONER_REFERENCE
        fhir_practitioner_role.practitioner = practitioner_reference
        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/practitioner-role-substitution-reference"
        reference = EnrolmentOfficerPractitionerConverter.build_fhir_resource_reference(
            self.test_substitution_officer,
        )
        extension.valueReference = reference
        fhir_practitioner_role.extension = [extension]
        return fhir_practitioner_role

    def verify_fhir_instance(self, fhir_obj):
        return None #FIXME
        self.assertIn(str(self.test_village.uuid), fhir_obj.location[0].reference)
        for identifier in fhir_obj.identifier:
            self.assertTrue(isinstance(identifier, Identifier))
            code = EnrolmentOfficerPractitionerRoleConverter.get_first_coding_from_codeable_concept(
                identifier.type).code
            if code == R4IdentifierConfig.get_fhir_generic_type_code():
                self.assertEqual(self.test_officer.code, identifier.value)
            elif code == R4IdentifierConfig.get_fhir_uuid_type_code():
                self.assertEqual(str(self.test_officer.uuid), identifier.value)
        self.assertIn(str(self.test_officer.uuid), fhir_obj.practitioner.reference)
        self.assertIn(str(self.test_substitution_officer.uuid),
                      fhir_obj.extension[0].valueReference.reference.split('Practitioner/')[1])
        self.assertEqual(1, len(fhir_obj.code))
        self.assertEqual(1, len(fhir_obj.code[0].coding))
        self.assertEqual("EO", fhir_obj.code[0].coding[0].code)
        self.assertEqual("Enrolment Officer", fhir_obj.code[0].coding[0].display)



from rest_framework import mixins
from rest_framework.viewsets import GenericViewSet

from api_fhir_r4.permissions import FHIRApiCoverageEligibilityRequestPermissions
from api_fhir_r4.serializers import CoverageEligibilityRequestSerializer
from api_fhir_r4.views.fhir.base import BaseFHIRView
from api_fhir_r4.views.filters import ValidityFromRequestParameterFilter
from insuree.models import Insuree


class CoverageEligibilityRequestViewSet(BaseFHIRView, mixins.CreateModelMixin, GenericViewSet):
    queryset = Insuree.filter_queryset()
    serializer_class = CoverageEligibilityRequestSerializer
    permission_classes = (FHIRApiCoverageEligibilityRequestPermissions,)

    def get_queryset(self):
        queryset = Insuree.get_queryset(None, self.request.user)
        return ValidityFromRequestParameterFilter(self.request).filter_queryset(queryset)



import logging

from django.core.exceptions import ObjectDoesNotExist

from api_fhir_r4.apps import ApiFhirConfig
from api_fhir_r4.configurations import R4LocationConfig, R4InvoiceConfig, GeneralConfiguration
from api_fhir_r4.converters import PatientConverter, BillInvoiceConverter, InvoiceConverter, \
    HealthFacilityOrganisationConverter
from api_fhir_r4.mapping.invoiceMapping import InvoiceTypeMapping, BillTypeMapping
from api_fhir_r4.subscriptions.notificationManager import RestSubscriptionNotificationManager
from api_fhir_r4.subscriptions.subscriptionCriteriaFilter import SubscriptionCriteriaFilter
from core.service_signals import ServiceSignalBindType
from core.signals import bind_service_signal

from openIMIS.openimisapps import openimis_apps

logger = logging.getLogger('openIMIS')
imis_modules = openimis_apps()


def bind_service_signals():
    if 'insuree' in imis_modules and GeneralConfiguration.get_subscribe_insuree_signal():
        def on_insuree_create_or_update(**kwargs):
            try:
                model = kwargs.get('result', None)
                if model:
                    notify_subscribers(model, PatientConverter(), 'Patient', None)
            except Exception as e:
                logger.error("Error while processing Patient Subscription", exc_info=e)

        bind_service_signal(
            'insuree_service.create_or_update',
            on_insuree_create_or_update,
            bind_type=ServiceSignalBindType.AFTER
        )

    if 'location' in imis_modules and R4LocationConfig.get_subscribe_location_signal():
        def on_hf_create_or_update(**kwargs):
            try:
                model = kwargs.get('result', None)
                if model:
                    notify_subscribers(model, HealthFacilityOrganisationConverter(), 'Organisation', 'bus')
            except Exception as e:
                logger.error("Error while processing Organisation Subscription", exc_info=e)

        bind_service_signal(
            'health_facility_service.update_or_create',
            on_hf_create_or_update,
            bind_type=ServiceSignalBindType.AFTER
        )
    if 'invoice' in imis_modules and R4InvoiceConfig.get_subscribe_invoice_signal():
        from invoice.models import Bill, Invoice

        def on_bill_create(**kwargs):
            try:
                result = kwargs.get('result', {})
                if result and result.get('success', False):
                    model_uuid = result['data']['uuid']
                    model = Bill.objects.get(uuid=model_uuid)
                    notify_subscribers(model, BillInvoiceConverter(), 'Invoice',
                                       BillTypeMapping.invoice_type[model.subject_type.model])
            except Exception as e:
                logger.error("Error while processing Bill Subscription", exc_info=e)

        def on_invoice_create(**kwargs):
            try:
                result = kwargs.get('result', {})
                if result and result.get('success', False):
                    model_uuid = result['data']['uuid']
                    model = Invoice.objects.get(uuid=model_uuid)
                    notify_subscribers(model, InvoiceConverter(), 'Invoice',
                                       InvoiceTypeMapping.invoice_type[model.subject_type.model])
            except Exception as e:
                logger.error("Error while processing Invoice Subscription", exc_info=e)

        bind_service_signal(
            'signal_after_invoice_module_bill_create_service',
            on_bill_create,
            bind_type=ServiceSignalBindType.AFTER
        )
        bind_service_signal(
            'signal_after_invoice_module_invoice_create_service',
            on_invoice_create,
            bind_type=ServiceSignalBindType.AFTER
        )


def notify_subscribers(model, converter, resource_name, resource_type_name):
    try:
        subscriptions = SubscriptionCriteriaFilter(model, resource_name,
                                                   resource_type_name).get_filtered_subscriptions()
        RestSubscriptionNotificationManager(converter).notify_subscribers_with_resource(model, subscriptions)
    except Exception as e:
        logger.error(f'Notifying subscribers failed: {e}')
        import traceback
        logger.debug(traceback.format_exc())



from fhir.resources.R4B.subscription import Subscription

from api_fhir_r4.subscriptions import SubscriptionConverter
from api_fhir_r4.tests.mixin import ConvertToImisTestMixin, ConvertToFhirTestMixin, ConvertJsonToFhirTestMixin
from api_fhir_r4.tests.mixin.SubscriptionTestMixin import SubscriptionTestMixin
from api_fhir_r4.utils import TimeUtils


class SubscriptionConverterTestCase(SubscriptionTestMixin,
                                    ConvertToImisTestMixin,
                                    ConvertToFhirTestMixin,
                                    ConvertJsonToFhirTestMixin):
    converter = SubscriptionConverter
    fhir_resource = Subscription
    json_repr = 'test/test_subscription.json'

    def _load_json_repr(self):
        json_repr = super(SubscriptionConverterTestCase, self)._load_json_repr()
        json_repr['end'] = TimeUtils.str_iso_to_date(json_repr['end'])
        return json_repr





from fhir.resources.R4B.operationoutcome import OperationOutcome
from rest_framework import status

from api_fhir_r4.configurations import R4IssueTypeConfig


class FhirApiDeleteTestMixin(object):

    @property
    def base_url(self):
        raise NotImplementedError()

    @property
    def _test_request_data(self):
        raise NotImplementedError()

    def get_id_for_created_resource(self, response):
        raise NotImplementedError()

    def get_fhir_obj_from_json_response(self, response):
        raise NotImplementedError()

    def login(self):
        raise NotImplementedError()

    def test_delete_should_remove_correctly(self):
        self.login()
        # create
        self.create_dependencies()
        response = self.client.post(self.base_url, data=self._test_request_data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        resource_id = self.get_id_for_created_resource(response)
        # verify if exist
        response = self.client.get(self.base_url + resource_id + "/", data=None, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNotNone(response.content)

        response = self.client.delete(self.base_url + resource_id + "/", data=None, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertTrue(not response.content)

        # verify if deleted
        response = self.client.get(self.base_url + resource_id + "/", data=None, format='json')
        self.verify_deleted_response(response)

    def create_dependencies(self):
        pass

    def verify_deleted_response(self, response):
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        #FIXME detail is sent on the test iso OO
        # operation_outcome = self.get_fhir_obj_from_json_response(response)
        #self.assertTrue(isinstance(operation_outcome, OperationOutcome))
        #self.assertEqual(R4IssueTypeConfig.get_fhir_code_for_not_found(), operation_outcome.issue[0].code)



import json
import os

from django.utils.translation import gettext as _
from rest_framework.test import APITestCase
from rest_framework import status

from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.tests import (
    GenericFhirAPITestMixin,
    FhirApiReadTestMixin
)
from api_fhir_r4.tests.mixin.logInMixin import LogInMixin
from invoice.models import (
    Invoice,
    PaymentInvoice,
    DetailPaymentInvoice
)
from invoice.tests.helpers import create_test_invoice


class PaymentNoticeAPITests(GenericFhirAPITestMixin, FhirApiReadTestMixin, APITestCase, LogInMixin):

    base_url = GeneralConfiguration.get_base_url()+'PaymentNotice/'
    _test_json_path = "/test/test_payment_notice.json"
    _test_invoice_uuid = "bd84b2f2-ec1d-48de-8f8c-5a477aa4a29f"
    _test_json_path_credentials = "/test/test_login.json"
    _test_request_data_credentials = None
    _user = None

    def setUp(self):
        super(PaymentNoticeAPITests, self).setUp()
        self._user = self.get_or_create_user_api()

    def create_dependencies(self):
        invoice = create_test_invoice(subject=None, thirdparty=None, user=self._user, **{
            'id': self._test_invoice_uuid,
            'amount_total': '10000.0',
            'amount_net': '10000.0',
            'user_created': self._user,
            'user_updated': self._user
        })
        return invoice

    def _update_payload_wrong_status(self, data):
        data["status"] = 'activeee'
        return data

    def _update_payload_wrong_payment_status(self, data):
        data["paymentStatus"]['coding'][0]['code'] = 'cleatedddd'
        return data

    def _update_payload_wrong_request_type(self, data):
        data["request"]['type'] = 'Invoices'
        return data

    def test_post_should_create_correctly(self):
        invoice = self.create_dependencies()
        headers = self.initialize_auth()
        response = self.client.post(self.base_url, data=self._test_request_data, format='json', **headers)
        if response.status_code == status.HTTP_201_CREATED:
            response_json_payment = response.json()
            payment_id = response_json_payment['id']
            DetailPaymentInvoice.objects.filter(payment__id=payment_id).delete()
            PaymentInvoice.objects.filter(id=payment_id).delete()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertIsNotNone(response.content)
        Invoice.objects.filter(id=invoice.id).delete()

    def test_post_wrong_status(self):
        invoice = self.create_dependencies()
        headers = self.initialize_auth()
        modified_payload = self._update_payload_wrong_status(data=self._test_request_data)
        response = self.client.post(self.base_url, data=modified_payload, format='json', **headers)
        response_json_no_name = response.json()
        self.assertEqual(
            self.get_response_details(response_json_no_name),
            _("The request cannot be processed due to the following issues:\nBad value "
              "in paymentStatus, should be either 'active', 'cancelled', draft', 'entered-in-error'"
              ",\nBad value in paymentStatus, ""should be either 'active', 'cancelled', draft', 'entered-in-error'")
        )
        Invoice.objects.filter(id=invoice.id).delete()

    def test_post_wrong_payment_status(self):
        invoice = self.create_dependencies()
        headers = self.initialize_auth()
        modified_payload = self._update_payload_wrong_payment_status(data=self._test_request_data)
        response = self.client.post(self.base_url, data=modified_payload, format='json', **headers)
        response_json_no_name = response.json()
        self.assertEqual(
            self.get_response_details(response_json_no_name),
            _("The request cannot be processed due to the following "
              "issues:\nBad value in paymentStatus, should be either 'paid' or 'cleared'")
        )
        Invoice.objects.filter(id=invoice.id).delete()

    def test_post_wrong_request_type(self):
        invoice = self.create_dependencies()
        headers = self.initialize_auth()
        modified_payload = self._update_payload_wrong_request_type(data=self._test_request_data)
        response = self.client.post(self.base_url, data=modified_payload, format='json', **headers)
        response_json_no_name = response.json()
        self.assertEqual(
            self.get_response_details(response_json_no_name),
            _("The request cannot be processed due to the following "
              "issues:\nBad value in type in request, should be either 'bill' or 'invoice'")
        )
        Invoice.objects.filter(id=invoice.id).delete()



import os

from fhir.resources.R4B.invoice import Invoice

from api_fhir_r4.converters import BillInvoiceConverter
from api_fhir_r4.tests.mixin import ConvertToFhirTestMixin
from api_fhir_r4.tests.mixin.billInvoiceTestMixin import BillInvoiceTestMixin


class InvoiceConverterInvoiceTestCase(BillInvoiceTestMixin, ConvertToFhirTestMixin):
    converter = BillInvoiceConverter
    fhir_resource = Invoice
    json_repr = 'test/test_invoice.json'

    def setUp(self):
        super(InvoiceConverterInvoiceTestCase, self).setUp()

    def test_to_imis_obj(self):
        self.assertRaises(NotImplementedError, self.converter.to_imis_obj, object(), 1)



import json
import os

from rest_framework import status
from rest_framework.test import APITestCase

from api_fhir_r4.configurations import GeneralConfiguration, R4CommunicationRequestConfig as Config
from api_fhir_r4.tests import GenericFhirAPITestMixin
from api_fhir_r4.tests import LocationTestMixin
from api_fhir_r4.tests.utils import load_and_replace_json

from api_fhir_r4.tests.mixin.logInMixin import LogInMixin
from api_fhir_r4.utils import TimeUtils
from claim.models import Claim, ClaimItem, ClaimService, Feedback
from claim.test_helpers import create_test_claim_admin
from core import datetime
from insuree.test_helpers import create_test_insuree
from location.models import HealthFacility
from medical.models import Diagnosis
from medical.test_helpers import create_test_item, create_test_service
from location.test_helpers import create_test_village, create_test_health_facility


class CommunicationAPITests(GenericFhirAPITestMixin, APITestCase, LogInMixin):
    base_url = GeneralConfiguration.get_base_url() + 'Communication/'
    _test_json_path = "/test/test_communication.json"

    _test_json_path_credentials = "/test/test_login.json"
    _test_request_data_credentials = None
    _test_json_path_with_code_reference = "/test/test_communication_with_code_reference.json"

    # feedback expected data
    _TEST_FEE_UUID = "612a1e12-ce44-4632-90a8-129ec714ec59"
    _TEST_CARE_RENDERED = False
    _TEST_PAYMENT_ASKED = True
    _TEST_DRUG_PRESCRIBED = True
    _TEST_DRUG_RECEIVED = False
    _TEST_ASESSMENT = '3'

    # claim data
    _TEST_CLAIM_CODE = 'codeTest'
    _TEST_CLAIM_UUID = "7ac646cb-d3cd-4660-baeb-ee34ecf0354e"
    _TEST_STATUS = Claim.STATUS_ENTERED
    _TEST_STATUS_DISPLAY = "entered"
    _TEST_OUTCOME = "queued"
    _TEST_ADJUSTMENT = "adjustment"
    _TEST_DATE_PROCESSED = "2010-11-16T00:00:00"
    _TEST_APPROVED = 1000.00
    _TEST_REJECTION_REASON = 0
    _TEST_VISIT_TYPE = "O"

    # claim item data
    _TEST_ITEM_CODE = "iCode"
    _TEST_ITEM_UUID = "e2bc1546-390b-4d41-8571-632ecf7a0936"
    _TEST_ITEM_STATUS = Claim.STATUS_ENTERED
    _TEST_ITEM_QUANTITY = 20
    _TEST_ITEM_PRICE = 10.0
    _TEST_ITEM_REJECTED_REASON = 0

    # claim service data
    _TEST_SERVICE_CODE = "sCode"
    _TEST_SERVICE_UUID = "a17602f4-e9ff-4f42-a6a4-ccefcb23b4d6"
    _TEST_SERVICE_STATUS = Claim.STATUS_ENTERED
    _TEST_SERVICE_QUANTITY = 1
    _TEST_SERVICE_PRICE = 800
    _TEST_SERVICE_REJECTED_REASON = 0

    _TEST_ID = 9999
    _PRICE_ASKED = 1000
    _PRICE_APPROVED = 1000
    _ADMIN_AUDIT_USER_ID = 1

    _TEST_ITEM_AVAILABILITY = True

    _TEST_ITEM_TYPE = 'D'
    _TEST_SERVICE_TYPE = 'D'

    # insuree and claim admin data
    _TEST_INSUREE_UUID = "76aca309-f8cf-4890-8f2e-b416d78de00b"
    _TEST_INSUREE_CHFID = "999000001"
    _TEST_CLAIM_ADMIN_UUID = "044c33d1-dbf3-4d6a-9924-3797b461e535"

    # hf test data
    _TEST_HF_ID = 10000
    _TEST_HF_UUID = "6d0eea8c-62eb-11ea-94d6-c36229a16c2f"
    _TEST_HF_CODE = "12345678"
    _TEST_HF_NAME = "TEST_NAME"
    _TEST_HF_LEVEL = "H"
    _TEST_HF_LEGAL_FORM = "G"
    _TEST_ADDRESS = "TEST_ADDRESS"
    _TEST_PHONE = "133-996-476"
    _TEST_FAX = "1-408-999 8888"
    _TEST_EMAIL = "TEST@TEST.com"
    sub_str = {}
    test_insuree = None
    test_claim_admin = None
    test_hf = None
    test_village = None
    test_claim = None
    test_item = None
    test_service = None
    def setUp(self):
        super(CommunicationAPITests, self).setUp()
        self._test_request_data=load_and_replace_json(self._test_json_path,self.sub_str)
        self.get_or_create_user_api()
        self.test_insuree = create_test_insuree()
        self.test_village = self.test_insuree.current_village or self.test_insuree.family.location
        self.test_hf = self.create_test_hf()
        self.test_claim_admin = create_test_claim_admin( custom_props={'health_facility_id': self.test_hf.id})
        self.test_claim = self.create_test_claim()
        self.test_item = self.create_test_claim_item()
        self.test_service = self.create_test_claim_service()

        self.sub_str[self._TEST_INSUREE_UUID]=self.test_insuree.uuid
        self.sub_str[self._TEST_INSUREE_CHFID]=self.test_insuree.chf_id
        self.sub_str[self._TEST_CLAIM_ADMIN_UUID]=self.test_claim_admin.uuid
        self.sub_str[self._TEST_CLAIM_UUID]=self.test_claim.uuid
        self.sub_str[self._TEST_HF_UUID]=self.test_hf.uuid
        self._TEST_HF_UUID=self.test_hf.uuid
        self._TEST_HF_ID=self.test_hf.id



    def create_test_claim_item(self):
        item = ClaimItem()
        item.item = create_test_item(
            self._TEST_ITEM_TYPE,
            custom_props={"code": self._TEST_ITEM_CODE}
        )
        item.claim = self.test_claim
        item.status = self._TEST_ITEM_STATUS
        item.qty_approved = self._TEST_ITEM_QUANTITY
        item.qty_provided = self._TEST_ITEM_QUANTITY
        item.rejection_reason = self._TEST_ITEM_REJECTED_REASON
        item.availability = self._TEST_ITEM_AVAILABILITY
        item.price_asked = self._TEST_ITEM_PRICE
        item.price_approved = self._TEST_ITEM_PRICE
        item.audit_user_id = self._ADMIN_AUDIT_USER_ID
        item.save()
        return item

    def create_test_claim_service(self):
        service = ClaimService()
        service.service = create_test_service(
            self._TEST_SERVICE_TYPE,
            custom_props={"code": self._TEST_SERVICE_CODE}
        )
        service.claim = self.test_claim
        service.status = self._TEST_SERVICE_STATUS
        service.qty_approved = self._TEST_SERVICE_QUANTITY
        service.qty_provided = self._TEST_SERVICE_QUANTITY
        service.rejection_reason = self._TEST_SERVICE_REJECTED_REASON
        service.availability = self._TEST_ITEM_AVAILABILITY
        service.price_asked = self._TEST_SERVICE_PRICE
        service.price_approved = self._TEST_SERVICE_PRICE
        service.audit_user_id = self._ADMIN_AUDIT_USER_ID
        service.save()
        return service

    def create_test_hf(self):
        hf = create_test_health_facility(
            self._TEST_HF_CODE,
            self.test_village.parent.parent.id,
            custom_props = {
                'name': self._TEST_HF_NAME,
                'level':self._TEST_HF_LEVEL,
                'legal_form_id':self._TEST_HF_LEGAL_FORM,
                'address':self._TEST_ADDRESS,
                'phone':self._TEST_PHONE,
                'fax':self._TEST_FAX,
                'email':self._TEST_EMAIL,
            }
        )
        return hf

    def create_test_claim(self):
        imis_claim = Claim()
        imis_claim.id = self._TEST_ID
        imis_claim.code = self._TEST_CLAIM_CODE
        imis_claim.status = self._TEST_STATUS
        imis_claim.adjustment = self._TEST_ADJUSTMENT
        imis_claim.date_processed = TimeUtils.str_to_date(self._TEST_DATE_PROCESSED)
        imis_claim.approved = self._TEST_APPROVED
        imis_claim.rejection_reason = self._TEST_REJECTION_REASON
        imis_claim.insuree = self.test_insuree
        imis_claim.health_facility = self.test_hf
        imis_claim.icd = Diagnosis(code='ICD00I', name="test icd")
        imis_claim.icd.audit_user_id = self._ADMIN_AUDIT_USER_ID
        imis_claim.icd.save()
        imis_claim.audit_user_id = self._ADMIN_AUDIT_USER_ID
        imis_claim.icd.date_from = datetime.date(2018, 12, 12)
        imis_claim.date_from = datetime.date(2018, 12, 12)
        imis_claim.date_claimed = datetime.date(2018, 12, 14)
        imis_claim.visit_type = self._TEST_VISIT_TYPE
        imis_claim.admin = self.test_claim_admin
        imis_claim.save()
        return imis_claim



    def _get_json_of_communication_with_code_reference(self):
        dir_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        return json.loads(
            open(dir_path + self._test_json_path_with_code_reference).read()
            )

    def test_post_should_create_correctly(self):
        
        response = self.client.post(
            GeneralConfiguration.get_base_url() + 'login/', data=self._test_request_data_credentials, format='json'
        )
        response_json = response.json()
        token = response_json["token"]
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {token}"
        }


        dataset = [
            load_and_replace_json(self._test_json_path, self.sub_str),
            load_and_replace_json(self._test_json_path_with_code_reference, self.sub_str),
        ]

        for data in dataset:
            response = self.client.post(self.base_url, data=data, format='json', **headers)

            if False:#FIXME static data gets feedback exists, 
                self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())
                self.assertIsNotNone(response.content)
                response_json = response.json()
                self.assertEqual(len(response_json['payload']), 5)
                for payload in response_json['payload']:
                    code = payload['extension'][0]['valueCodeableConcept']['coding'][0]['code']
                    content_string = payload['contentString']
                    if code != Config.get_fhir_asessment_code():
                        bool_value = self._convert_bool_value(content_string)
                    if code == Config.get_fhir_care_rendered_code():
                        self.assertEqual(self._TEST_CARE_RENDERED, bool_value, f'code {code}: {content_string}')
                    elif code == Config.get_fhir_payment_asked_code():
                        self.assertEqual(self._TEST_PAYMENT_ASKED, bool_value, f'code {code}: {content_string}')
                    elif code == Config.get_fhir_drug_prescribed_code():
                        self.assertEqual(self._TEST_DRUG_PRESCRIBED, bool_value, f'code {code}: {content_string}')
                    elif code == Config.get_fhir_drug_received_code():
                        self.assertEqual(self._TEST_DRUG_RECEIVED, bool_value, f'code {code}: {content_string}')
                    elif code == Config.get_fhir_asessment_code():
                        self.assertEqual(self._TEST_ASESSMENT, content_string, f'code {code}: {content_string}')

        
        claim = Claim.objects.get(uuid=str(self.test_claim.uuid))
        self.assertEqual(claim.feedback_status, Claim.FEEDBACK_DELIVERED)
        self.assertTrue(claim.feedback_available)
        self.assertIsNotNone(claim.feedback)
        if False:#FIXME
            self.assertEqual(claim.feedback.uuid.lower(), response_json['identifier'][0]['value'].lower())
            self.assertEqual(claim.uuid.lower(), response_json['about'][0]['identifier']['value'].lower())



    def _convert_bool_value(self, fhir_content_string):
        if fhir_content_string == "yes":
            return True
        if fhir_content_string == "no":
            return False
        return None



from fhir.resources.R4B.paymentnotice import PaymentNotice

from api_fhir_r4.paymentNotice import PaymentNoticeConverter
from api_fhir_r4.tests.mixin import ConvertJsonToFhirTestMixin
from api_fhir_r4.tests.mixin.logInMixin import LogInMixin
from api_fhir_r4.tests.mixin.paymentNoticeTestMixin import PaymentNoticeTestMixin
from invoice.models import (
    Invoice,
    PaymentInvoice,
    DetailPaymentInvoice
)
from invoice.tests.helpers import create_test_invoice


class PaymentNoticeConverterTestCase(PaymentNoticeTestMixin, LogInMixin, ConvertJsonToFhirTestMixin):
    _TEST_USER_NAME = "TestUserTest2"

    converter = PaymentNoticeConverter
    fhir_resource = PaymentNotice
    json_repr = 'test/test_payment_notice.json'

    def create_dependencies(self, user):
        invoice = create_test_invoice(subject=None, thirdparty=None, user=user, **{
            'id': 'bd84b2f2-ec1d-48de-8f8c-5a477aa4a29f',
            'amount_total': '10000.0',
            'amount_net': '10000.0',
            'user_created': user,
            'user_updated': user,
            'date_created': '2020-01-01'
        })
        return invoice

    def test_to_fhir_obj(self):
        user = self.get_or_create_user_api()
        invoice = self.create_dependencies(user)
        imis_payment, imis_payment_detail = self.create_test_imis_instance()
        imis_payment.save(username=user.username)
        imis_payment_detail.payment = imis_payment
        imis_payment_detail.subject = invoice
        imis_payment_detail.save(username=user.username)
        fhir_payment_notice = self.converter.to_fhir_obj(imis_payment)
        self.verify_fhir_instance(fhir_payment_notice)
        self._delete_created_entries(imis_payment, imis_payment_detail, invoice)

    def test_to_imis_obj(self):
        user = self.get_or_create_user_api()
        invoice = self.create_dependencies(user)
        fhir_payment_notice = self.create_test_fhir_instance()
        imis_payment = PaymentNoticeConverter.to_imis_obj(fhir_payment_notice)
        imis_payment.save(username=user.username)
        imis_payment_detail = imis_payment.imis_payment_detail
        imis_payment_detail.payment = imis_payment
        imis_payment_detail.subject = invoice
        if 'reconciliation' in imis_payment.json_ext:
            imis_payment_detail.reconcilation_id = imis_payment.json_ext['reconciliation']['id']
        imis_payment_detail.save(username=user.username)
        imis_invoice_status = imis_payment.invoice_status
        self.verify_imis_instance(imis_payment)
        self.verify_imis_detail_instance(imis_payment, imis_payment_detail)
        self.verify_imis_invoice_status(imis_invoice_status)
        self._delete_created_entries(imis_payment, imis_payment_detail, invoice)

    def _delete_created_entries(self, payment, payment_detail, invoice):
        DetailPaymentInvoice.objects.filter(id=payment_detail.id).delete()
        PaymentInvoice.objects.filter(id=payment.id).delete()
        Invoice.objects.filter(id=invoice.id).delete()



from rest_framework.request import Request

from api_fhir_r4.mapping.invoiceMapping import InvoiceTypeMapping, BillTypeMapping
from api_fhir_r4.mixins import (
    MultiIdentifierRetrieveManySerializersMixin,
    MultiIdentifierRetrieverMixin
)
from api_fhir_r4.model_retrievers import (
    CodeIdentifierModelRetriever,
    DatabaseIdentifierModelRetriever,
    UUIDIdentifierModelRetriever
)
from api_fhir_r4.multiserializer import modelViewset
from api_fhir_r4.permissions import (
    FHIRApiInvoicePermissions,
    FHIRApiBillPermissions
)
from api_fhir_r4.serializers import (
    InvoiceSerializer,
    BillSerializer
)
from api_fhir_r4.views.fhir.base import BaseMultiserializerFHIRView
from api_fhir_r4.views.filters import DateUpdatedRequestParameterFilter
from invoice.models import Bill
from invoice.models import Invoice


class InvoiceViewSet(BaseMultiserializerFHIRView,
                     modelViewset.MultiSerializerModelViewSet,
                     MultiIdentifierRetrieveManySerializersMixin, MultiIdentifierRetrieverMixin):
    retrievers = [UUIDIdentifierModelRetriever, DatabaseIdentifierModelRetriever, CodeIdentifierModelRetriever]
    lookup_field = 'identifier'

    @property
    def serializers(self):
        return {
            InvoiceSerializer:
                (self._invoice_queryset(), self._invoice_serializer_validator, (FHIRApiInvoicePermissions,)),
            BillSerializer:
                (self._bill_queryset(), self._bill_serializer_validator, (FHIRApiBillPermissions,))
        }

    @classmethod
    def _invoice_serializer_validator(cls, context):
        return cls._base_request_validator_dispatcher(
            request=context['request'],
            get_check=lambda x: cls._get_type_from_query(x) in ('invoice', None),
            post_check=lambda x: cls._get_type_from_body(x) in [item['code'] for item in
                                                                InvoiceTypeMapping.invoice_type],
            put_check=lambda x: cls._get_type_from_body(x) in [item['code'] for item in
                                                               InvoiceTypeMapping.invoice_type],
        )

    @classmethod
    def _bill_serializer_validator(cls, context):
        return cls._base_request_validator_dispatcher(
            request=context['request'],
            get_check=lambda x: cls._get_type_from_query(x) in ('bill', None),
            post_check=lambda x: cls._get_type_from_body(x) in [item['code'] for item in BillTypeMapping.invoice_type],
            put_check=lambda x: cls._get_type_from_body(x) in [item['code'] for item in BillTypeMapping.invoice_type],
        )

    @classmethod
    def _base_request_validator_dispatcher(cls, request: Request, get_check, post_check, put_check):
        if request.method == 'GET':
            return get_check(request)
        elif request.method == 'POST':
            return post_check(request)
        elif request.method == 'PUT':
            return put_check(request)
        return True

    def list(self, request, *args, **kwargs):
        identifier = request.GET.get("code")
        if identifier:
            return self.retrieve(request, *args, **{**kwargs, 'identifier': identifier})
        return super().list(request, *args, **kwargs)

    def get_queryset(self):
        return Invoice.objects

    def _invoice_queryset(self):
        queryset = Invoice.objects.filter(is_deleted=False).order_by('date_created')
        return DateUpdatedRequestParameterFilter(self.request).filter_queryset(queryset)

    def _bill_queryset(self):
        queryset = Bill.objects.filter(is_deleted=False).order_by('date_created')
        return DateUpdatedRequestParameterFilter(self.request).filter_queryset(queryset)

    @classmethod
    def _get_type_from_body(cls, request):
        try:
            return request.data['type'][0]['coding'][0]['code'].lower()
        except KeyError:
            return None

    @classmethod
    def _get_type_from_query(cls, request):
        try:
            return request.GET['resourceType'].lower()
        except KeyError:
            return None



# Generated by Django 3.0.14 on 2022-03-13 16:34

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('api_fhir_r4', '0002_auto_20220313_1607'),
    ]

    operations = [
        migrations.AlterField(
            model_name='subscriptionnotificationresult',
            name='id',
            field=models.UUIDField(db_column='UUID', default=uuid.uuid4, editable=False, primary_key=True, serialize=False),
        ),
    ]



import json
import os

from core.models import User
from fhir.resources.R4B import construct_fhir_element
from rest_framework import status

from api_fhir_r4.configurations import (
    R4IdentifierConfig,
    GeneralConfiguration
)
from api_fhir_r4.converters import BaseFHIRConverter
from fhir.resources.R4B.bundle import Bundle
from api_fhir_r4.utils import DbManagerUtils


class GenericFhirAPITestMixin(object):

    @property
    def base_url(self):
        return None

    @property
    def _test_json_path(self):
        return None
    
    @property
    def _test_json_path_credentials(self):
        return None
    
    _TEST_SUPERUSER_NAME = 'admin'
    _TEST_SUPERUSER_PASS = 'adminadmin'#'Admin123'
    _test_request_data = None
    _test_json_path_credentials = None

    def setUp(self):
        dir_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        if self._test_json_path and  self._test_request_data is None:
            json_representation = open(dir_path + self._test_json_path).read()
            self._test_request_data = json.loads(json_representation)
        if self._test_json_path_credentials and  self._test_request_data_credentials is None:
            json_representation = open(dir_path + self._test_json_path_credentials).read()
            self._test_request_data_credentials = json.loads(json_representation)

    def apply_replace_map(self , payload):
        return payload

    def initialize_auth(self):
        response = self.client.post(
            GeneralConfiguration.get_base_url() + 'login/', data=self._test_request_data_credentials, format='json'
        )
        response_json = response.json()
        token = response_json["token"]
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {token}"
        }
        return headers
    def get_response_details(self, response_json):
        if 'issue' in response_json\
            and len(response_json["issue"])>0\
            and 'details' in response_json["issue"][0]\
            and 'text' in response_json["issue"][0]["details"]: 
            return response_json["issue"][0]["details"]['text'] 

        
        elif 'detail' in response_json:
            return response_json["detail"]
  
    def login(self):
        user = DbManagerUtils.get_object_or_none(User, username=self._TEST_SUPERUSER_NAME)
        if user is None:
            user = self.__create_superuser()
        self.client.force_authenticate(user=user)

    def __create_superuser(self):
        User.objects.create_superuser(username=self._TEST_SUPERUSER_NAME, password=self._TEST_SUPERUSER_PASS)
        return DbManagerUtils.get_object_or_none(User, username=self._TEST_SUPERUSER_NAME)

    def get_bundle_from_json_response(self, response):
        response_json = response.json()
        bundle = construct_fhir_element(response_json['resourceType'], response_json)
        self.assertTrue(isinstance(bundle, Bundle))
        return bundle

    def get_id_for_created_resource(self, response):
        result = None
        response_json = response.json()
        fhir_obj = construct_fhir_element(response_json['resourceType'], response_json)
        if hasattr(fhir_obj, 'identifier'):
            result = BaseFHIRConverter.get_fhir_identifier_by_code(fhir_obj.identifier,
                                                                   R4IdentifierConfig.get_fhir_uuid_type_code())
        return result

    def get_fhir_obj_from_json_response(self, response):
        fhir_obj = None
        response_json = response.json()
        if 'resourceType' in response_json:
            fhir_obj = construct_fhir_element(response_json['resourceType'], response_json)
            
        return fhir_obj

    def test_get_should_required_login(self):
        response = self.client.get(self.base_url, data=None, format='json')
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)



from rest_framework.authentication import SessionAuthentication
from core.jwt import *
from api_fhir_r4.views.login_viewset import LoginView


class CsrfExemptSessionAuthentication(SessionAuthentication):
    def enforce_csrf(self, request):
        return



from api_fhir_r4.serializers import CodeSystemSerializer

from rest_framework import viewsets
from rest_framework.response import Response
from rest_framework.permissions import IsAuthenticated

from rest_framework.views import APIView
from api_fhir_r4.views import CsrfExemptSessionAuthentication


class CodeSystemOrganizationHFLegalFormViewSet(viewsets.ViewSet):
    serializer_class = CodeSystemSerializer
    permission_classes = (IsAuthenticated,)
    authentication_classes = [CsrfExemptSessionAuthentication] + APIView.settings.DEFAULT_AUTHENTICATION_CLASSES

    def list(self, request):
        # we don't use typical instance, we only indicate the model and the field to be mapped into CodeSystem
        serializer = CodeSystemSerializer(
            instance=None,
            **{
                'model_name': 'HealthFacilityLegalForm',
                'code_field': 'code',
                'display_field': 'legal_form',
                'id': 'organization-hf-legal-form',
                'name': 'OrganizationHFLegalFormCS',
                'title': 'Legal Forms (Organization)',
                'description': 'Indicates the legal forms of the Organization. '
                               'Values defined by openIMIS. Can be extended.',
                'url': self.request.build_absolute_uri()
            }
        )
        data = serializer.to_representation(obj=None)
        return Response(data)



from api_fhir_r4.configurations import ClaimConfiguration


class R4ClaimConfig(ClaimConfiguration):

    @classmethod
    def build_configuration(cls, cfg):
        cls.get_config().R4_fhir_claim_config = cfg['R4_fhir_claim_config']

    @classmethod
    def get_fhir_claim_information_guarantee_id_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_information_guarantee_id_code',
                                                                    "guarantee")

    @classmethod
    def get_fhir_claim_information_explanation_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_information_explanation_code',
                                                                    "explanation")

    @classmethod
    def get_fhir_claim_item_explanation_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_item_explanation_code',
                                                                    "item_explanation")

    @classmethod
    def get_fhir_claim_item_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_item_code', "item")

    @classmethod
    def get_fhir_claim_service_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_service_code', "service")

    @classmethod
    def get_fhir_claim_status_rejected_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_status_rejected_code', "rejected")

    @classmethod
    def get_fhir_claim_status_entered_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_status_entered_code', "entered")

    @classmethod
    def get_fhir_claim_status_checked_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_status_checked_code', "checked")

    @classmethod
    def get_fhir_claim_status_processed_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_status_processed_code', "processed")

    @classmethod
    def get_fhir_claim_status_valuated_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_status_valuated_code', "valuated")

    @classmethod
    def get_fhir_claim_item_status_passed_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_item_status_passed_code', "passed")

    @classmethod
    def get_fhir_claim_item_status_rejected_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_item_status_rejected_code', "rejected")

    @classmethod
    def get_fhir_claim_item_general_adjudication_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_item_general_adjudication_code',
                                                                    "general")

    @classmethod
    def get_fhir_claim_item_rejected_reason_adjudication_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_item_rejected_reason_adjudication_code',
                                                                    "rejected_reason")

    @classmethod
    def get_fhir_claim_organization_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_organization_code',
                                                                    "openIMIS")  # has to be updated when 'Organization' is created

    @classmethod
    def get_fhir_claim_attachment_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_attachment_code', "attachment")

    @classmethod
    def get_fhir_claim_visit_type_system(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_visit_type_system',
                                                                    'CodeSystem/claim-visit-type')

    @classmethod
    def get_fhir_claim_supporting_info_system(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_supporting_info_category_system',
                                                                    'CodeSystem/claim-supporting-info-category')

    @classmethod
    def get_fhir_item_reference_extension_system(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_item_reference_extension_system',
                                                                    'StructureDefinition/claim-item-reference')

    @classmethod
    def get_fhir_claim_item_category_system(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_item_category_system',
                                                                    'CodeSystem/claim-item-category')

    @classmethod
    def get_fhir_claim_diagnosis_system(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get('fhir_claim_diagnosis_system',
                                                                    'CodeSystem/diagnosis-ICD10-level1')

    @classmethod
    def get_fhir_claim_attachment_system(cls):
        pass

    @classmethod
    def get_allowed_fhir_claim_attachment_mime_types_regex(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            'fhir_claim_allowed_mime_types_regex',
            '|'.join(['text\/.*', 'image\/png', 'image\/jpe?g', 'application\/msword', '.*doc.*'])
        )



from api_fhir_r4.tests import GenericTestMixin
from api_fhir_r4.utils import DbManagerUtils
from location.models import Location
from policyholder.models import PolicyHolder


class PolicyHolderOrganisationTestMixin(GenericTestMixin):
    _TEST_UUID = "1234-1234-1234-1234"
    _TEST_LEGAL_FORM = 2
    _TEST_ACTIVITY_CODE = 3
    _TEST_CODE = "ABCD"
    _TEST_TRADE_NAME = "Test test"
    _TEST_LOCATION_ID = 22
    _TEST_REGION = 'Tahida'
    _TEST_DISTRICT = 'Rajo'
    _TEST_MUNICIPALITY = 'Jaber'
    _TEST_VILLAGE = 'Utha'
    _TEST_CONTACT_NAME = "Name"
    _TEST_CONTACT_SURNAME = "Surname"
    _TEST_ADDRESS_LINE = "Test address 21"
    _TEST_EMAIL = "email@email.com"
    _TEST_PHONE = "111111111"
    _TEST_FAX = "123123123"
    _FIXED_TYPE = "bus"
    _FIXED_LOCATION_TYPE = "physical"

    def create_test_fhir_instance(self):
        NotImplementedError('PH Organization to_imis_obj() not implemented.')

    def verify_fhir_instance(self, fhir_obj):
        self.assertEqual(self._TEST_UUID, fhir_obj.id)

        actual_legal_form = next(iter(
            extension.valueCodeableConcept.coding[0].code for extension in fhir_obj.extension if
            "legal-form" in extension.url), None)
        self.assertEqual(str(self._TEST_LEGAL_FORM), actual_legal_form)

        actual_activity = next(iter(
            extension.valueCodeableConcept.coding[0].code for extension in fhir_obj.extension if
            "activity" in extension.url), None)
        self.assertEqual(str(self._TEST_ACTIVITY_CODE), actual_activity)

        actual_id_uuid = next(iter(
            identifier.value for identifier in fhir_obj.identifier if
            "UUID" == identifier.type.coding[0].code), None)
        self.assertEqual(self._TEST_UUID, actual_id_uuid)

        actual_id_code = next(iter(
            identifier.value for identifier in fhir_obj.identifier if
            "Code" == identifier.type.coding[0].code), None)
        self.assertEqual(self._TEST_CODE, actual_id_code)

        self.assertEqual(self._TEST_REGION, fhir_obj.address[0].state)
        self.assertEqual(self._TEST_DISTRICT, fhir_obj.address[0].district)
        self.assertEqual(self._TEST_VILLAGE, fhir_obj.address[0].city)
        self.assertEqual(self._TEST_ADDRESS_LINE, fhir_obj.address[0].line[0])

        self.assertEqual(self._FIXED_LOCATION_TYPE, fhir_obj.address[0].type)

        actual_municipality = next(iter(
            extension.valueString for extension in fhir_obj.address[0].extension if
            "municipality" in extension.url), None)
        self.assertEqual(self._TEST_MUNICIPALITY, actual_municipality)

        expected_contact_name = "%s %s" % (self._TEST_CONTACT_NAME, self._TEST_CONTACT_SURNAME)
        actual_contact_name = fhir_obj.contact[0]['name'].text
        self.assertEqual(expected_contact_name, actual_contact_name)

        actual_email = next(iter(telecom.value for telecom in fhir_obj.telecom if
                                 "email" == telecom.system), None)
        self.assertEqual(self._TEST_EMAIL, actual_email)

        actual_fax = next(iter(telecom.value for telecom in fhir_obj.telecom if
                               "fax" == telecom.system), None)
        self.assertEqual(self._TEST_FAX, actual_fax)

        actual_phone = next(iter(telecom.value for telecom in fhir_obj.telecom if
                                 "phone" == telecom.system), None)
        self.assertEqual(self._TEST_PHONE, actual_phone)
        self.assertEqual(self._FIXED_TYPE, fhir_obj.type[0].coding[0].code)

    def verify_imis_instance(self, imis_obj):
        NotImplementedError('PH Organization to_imis_obj() not implemented.')

    def create_test_imis_instance(self):
        location = DbManagerUtils.get_object_or_none(Location, id=self._TEST_LOCATION_ID)

        return PolicyHolder(**{
            "uuid": self._TEST_UUID,
            "legal_form": self._TEST_LEGAL_FORM,
            "activity_code": self._TEST_ACTIVITY_CODE,
            "code": self._TEST_CODE,
            "trade_name": self._TEST_TRADE_NAME,
            "locations": location,
            "contact_name": {"name": self._TEST_CONTACT_NAME, "surname": self._TEST_CONTACT_SURNAME},
            "address": {"address": self._TEST_ADDRESS_LINE},
            "email": self._TEST_EMAIL,
            "fax": self._TEST_FAX,
            "phone": self._TEST_PHONE,
        })



import copy

from location.gql_mutations import update_or_create_health_facility

from api_fhir_r4.converters import HealthFacilityOrganisationConverter
from api_fhir_r4.serializers import BaseFHIRSerializer


class HealthFacilityOrganisationSerializer(BaseFHIRSerializer):
    fhirConverter = HealthFacilityOrganisationConverter()

    def create(self, validated_data):
        data = copy.deepcopy(validated_data)

        # UUID is automatically generated for HF model, has to be removed as create_or_update will fail.
        data.pop('uuid', None)
        return self.__create_or_update(data)

    def update(self, instance, validated_data):
        validated_data['id'] = instance.id
        validated_data['uuid'] = instance.uuid

        data = copy.deepcopy(validated_data)
        return self.__create_or_update(data)

    def __create_or_update(self, validated_data):
        request = self.context.get("request")
        user = request.user
        data = copy.deepcopy(validated_data)
        if '_state' in validated_data:
            del data['_state']

        if not data.get('offline', None):
            data['offline'] = False

        return update_or_create_health_facility(data, user)



import copy
import json
import os

from policy.models import Policy
from rest_framework.status import is_success

from api_fhir_r4.tests.mixin.logInMixin import LogInMixin
from api_fhir_r4.utils import DbManagerUtils
from rest_framework.test import APITestCase
from rest_framework import status
from core.models import User
from core.services import create_or_update_interactive_user, create_or_update_core_user
from api_fhir_r4.tests import GenericFhirAPITestMixin
from api_fhir_r4.tests.utils import load_and_replace_json
from api_fhir_r4.configurations import GeneralConfiguration
from core.test_helpers import create_test_officer
from insuree.models import Insuree
from insuree.test_helpers import create_test_insuree
from product.test_helpers import create_test_product


class ContractAPITests(GenericFhirAPITestMixin, APITestCase, LogInMixin):

    base_url = GeneralConfiguration.get_base_url()+'Contract/'
    _test_json_path = "/test/test_contract.json"

    _TEST_GROUP_UUID = "e8bbb7e4-19ef-4bef-9342-9ab6b9a928d3"
    _TEST_OFFICER_UUID = "ff7db42d-874b-400a-bba7-e59b273ae123"
    _TEST_INSUREE_UUID = "f8c56ada-d76d-4f6c-aad3-cfddc9fb38eb"
    _TEST_PRODUCT_CODE = "TE123"
    _TEST_PRODUCT_UUID = "8ed8d2d9-2644-4d29-ba37-ab772386cfca"

    _test_json_path_credentials = "/test/test_login.json"
    _test_request_data_credentials = None
    test_insuree = None
    test_product = None
    test_officer = None
    sub_str={}
    def setUp(self):
        super(ContractAPITests, self).setUp()
        self.get_or_create_user_api()
# example of use
#        self.sub_str[self._TEST_VILLAGE_UUID] = self.test_village.uuid
#        self.sub_str[self._TEST_SUBSTITUTION_OFFICER_UUID] = self.test_officer.uuid
        self.sub_str[self._TEST_GROUP_UUID] = self.test_insuree.family.uuid
        self.sub_str[self._TEST_INSUREE_UUID] = self.test_insuree.uuid
#        self.sub_str[self._TEST_INSUREE_CHFID] = self.test_insuree.chf_id
        self.sub_str[self._TEST_OFFICER_UUID] = self.test_officer.uuid
#        self.sub_str[self._TEST_CLAIM_ADMIN_UUID] = self.test_claim_admin.uuid
        self.sub_str[self._TEST_PRODUCT_UUID] = self.test_product.uuid
#        
        self._test_request_data = load_and_replace_json(self._test_json_path,self.sub_str)
        
    @classmethod
    def setUpTestData(cls):
        # create mocked insuree
        cls.test_insuree = create_test_insuree(with_family=True)

        # create mocked product
        cls.test_product = create_test_product(cls._TEST_PRODUCT_CODE, valid=True)
   

        # create mocked officer
        cls.test_officer = create_test_officer(custom_props={'uuid':cls._TEST_OFFICER_UUID})


    def test_post_should_create_correctly(self):
        headers = self._build_headers()
        response = self.client.post(self.base_url, data=self._test_request_data, format='json', **headers)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())
        self.assertIsNotNone(response.content)

    def test_post_patient_subject_reference(self):
        # Test use case when instead of family, subject is referencing insuree without family
        headers = self._build_headers()

        test_request_data = copy.deepcopy(self._test_request_data)
        #test_request_data['subject'][0]['reference'] = F"Patient/{self.test_insuree.uuid}"
        #test_request_data['author']['reference'] = F"Practitioner/{self.test_officer.uuid}"
        response = self.client.post(self.base_url, data=test_request_data, format='json', **headers)

        self._assert_contract_created(response)
        self._assert_insuree_family_created_in_process(response)

    def _build_headers(self):
        response = self.client.post(
            path=GeneralConfiguration.get_base_url() + 'login/',
            data=self._test_request_data_credentials,
            format='json'
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK,
            f"Login request has failed, bearer token not received, details: {response.json()}")

        token = response.json().get("token")
        return {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {token}"
        }

    def _assert_contract_created(self, response):
        response_status, content = response.status_code, response.json()
        self.assertEqual(
            response_status, status.HTTP_201_CREATED,
            F"Contract POST should return response with code 201, it returned {response_status} instead. \n"
            F"Request details:\n{content}")

        self.assertIsNotNone(
            content,
            "POST Response is marked as success, but response content is empty.")

        policy_id = content['identifier'][0]['value']
        self.assertTrue(
            Policy.objects.filter(uuid__iexact=policy_id, validity_to__isnull=True).exists(),
            "POST Contract returned `201 Created`, but policy object not stored in database")

    def _assert_insuree_family_created_in_process(self, response):
        policy_id = response.json()['identifier'][0]['value']
        policy = Policy.objects.get(uuid=policy_id)
        insuree = Insuree.objects.get(uuid=self.test_insuree.uuid)
        family_policy = policy.family
        family_insuree = insuree.family

        self.assertEqual(
            family_policy, family_insuree,
            F"Contract for insuree reference created but insuree family `{family_insuree}` is "
            F"different from policy family `{family_policy}`.")

        self.assertEqual(
            insuree.current_village, family_policy.location,
            "Family created from insuree object should share location their location.")

        self.assertEqual(family_policy.head_insuree, insuree,
                         "Head of family created from contract is different from contract subject.")

        

from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.tests import GenericFhirAPITestMixin, FhirApiReadTestMixin
from rest_framework.test import APITestCase
from api_fhir_r4.tests.mixin.logInMixin import LogInMixin
from graphql_jwt.shortcuts import get_token
from core.test_helpers import create_test_interactive_user
from dataclasses import dataclass
from core.models import User
from rest_framework import status


@dataclass
class DummyContext:
    """ Just because we need a context to generate. """
    user: User

class ContractAPITests(GenericFhirAPITestMixin, FhirApiReadTestMixin, APITestCase, LogInMixin):
    base_url = GeneralConfiguration.get_base_url() + 'Contract/'
    _test_json_path = None
    _test_request_data_credentials = None
    admin_token = None 
    admin_user = None

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.admin_user = create_test_interactive_user(username="testLocationAdmin")
        cls.admin_token = get_token(cls.admin_user, DummyContext(user=cls.admin_user))


    def test_simple_list_page_2(self):
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {self.admin_token}"
        }
        response = self.client.get(self.base_url+ '?page-offset=2', format='json', **headers)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNotNone(response.content)


from claim.models import ClaimAdmin
from django.utils.translation import gettext as _
from api_fhir_r4.utils import TimeUtils
from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig
from api_fhir_r4.converters import ClaimAdminPractitionerConverter, HealthFacilityOrganisationConverter
from fhir.resources.R4B.contactpoint import ContactPoint
from fhir.resources.R4B.extension import Extension
from fhir.resources.R4B.humanname import HumanName
from fhir.resources.R4B.identifier import Identifier
from fhir.resources.R4B.practitioner import Practitioner, PractitionerQualification
from fhir.resources.R4B.reference import Reference

from api_fhir_r4.models.imisModelEnums import ContactPointSystem, ContactPointUse
from api_fhir_r4.tests import GenericTestMixin, LocationTestMixin
from api_fhir_r4.utils import TimeUtils
from location.models import HealthFacility
from location.test_helpers import create_test_village, create_test_health_facility
from claim.test_helpers import create_test_claim_admin

class ClaimAdminPractitionerTestMixin(GenericTestMixin):

    _TEST_CLAIM_ADMIN_LAST_NAME = "Smith"
    _TEST_CLAIM_ADMIN_OTHER_NAME = "John"
    _TEST_CLAIM_ADMIN_DOB = "1990-03-24"
    _TEST_CLAIM_ADMIN_ID = 1
    _TEST_CLAIM_ADMIN_UUID = "254f6268-964b-4d8d-aa26-20081f22235e"
    _TEST_CLAIM_ADMIN_CODE = "1234abcd"
    _TEST_CLAIM_ADMIN_PHONE = "813-996-476"
    _TEST_CLAIM_ADMIN_EMAIL = "TEST@TEST.com"
    _TEST_CLAIM_ADMIN_FAX = "1-408-999 8888"
    _TEST_CLAIM_ADMIN_ADDRESS = "TEST_ADDRESS"

    _TEST_HF_ID = 90000
    _TEST_HF_UUID = "3fd89df0-a532-4df6-9b77-3cdb33b883f2"
    _TEST_HF_CODE = "92345678"
    _TEST_HF_NAME = "TEST_NAME"
    _TEST_HF_LEVEL = "H"
    _TEST_HF_LEGAL_FORM ="G"
    
    
    test_claim_admin = None
    test_hf = None
    sub_str = {}
    
    def setUp(self):
        super(ClaimAdminPractitionerTestMixin, self).setUp()
        self.test_village = create_test_village()
        self.test_hf=self.create_test_hf()
        self.test_claim_admin = create_test_claim_admin( custom_props={
            'health_facility_id': self.test_hf.id, 
            'code':self._TEST_CLAIM_ADMIN_CODE,
            'dob':TimeUtils.str_to_date(self._TEST_CLAIM_ADMIN_DOB),
            'phone': self._TEST_CLAIM_ADMIN_PHONE,
            'email_id': self._TEST_CLAIM_ADMIN_EMAIL ,
            'last_name': self._TEST_CLAIM_ADMIN_LAST_NAME,
            'other_names': self._TEST_CLAIM_ADMIN_OTHER_NAME}
                                            )
        self.sub_str[self._TEST_HF_UUID]=self.test_hf.uuid
        self.sub_str[self._TEST_CLAIM_ADMIN_UUID]=self.test_claim_admin.uuid

    def create_test_hf(self):
        self.test_hf = create_test_health_facility(
            self._TEST_HF_CODE,
            self.test_village.parent.parent.id,
            custom_props = {
                'name': self._TEST_HF_NAME,
                'level':self._TEST_HF_LEVEL,
                'legal_form_id':self._TEST_HF_LEGAL_FORM,
            }
        )
        self._TEST_HF_ID = self.test_hf.id
        self._TEST_HF_UUID = self.test_hf.uuid
        return self.test_hf

    def create_test_imis_instance(self, location=None, hf = None):
        return self.test_claim_admin


    def verify_imis_instance(self, imis_obj):
        self.assertEqual(self._TEST_CLAIM_ADMIN_LAST_NAME, imis_obj.last_name)
        self.assertEqual(self._TEST_CLAIM_ADMIN_OTHER_NAME, imis_obj.other_names)
        self.assertEqual(self._TEST_CLAIM_ADMIN_CODE, imis_obj.code)
        self.assertEqual(self._TEST_CLAIM_ADMIN_DOB+"T00:00:00", imis_obj.dob.isoformat())
        self.assertEqual(self._TEST_CLAIM_ADMIN_PHONE, imis_obj.phone)
        self.assertEqual(self._TEST_CLAIM_ADMIN_EMAIL, imis_obj.email_id)
        # we are not checking the extension as it's optional and the health facility
        # membership should be changed otherwise

    def create_test_fhir_instance(self):
        fhir_practitioner = Practitioner.construct()
        name = HumanName.construct()
        name.family = self._TEST_CLAIM_ADMIN_LAST_NAME
        name.given = [self._TEST_CLAIM_ADMIN_OTHER_NAME]
        name.use = "usual"
        fhir_practitioner.name = [name]
        identifiers = []
        code = ClaimAdminPractitionerConverter.build_fhir_identifier(
            self._TEST_CLAIM_ADMIN_CODE,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            R4IdentifierConfig.get_fhir_generic_type_code()
        )
        identifiers.append(code)
        fhir_practitioner.identifier = identifiers
        fhir_practitioner.birthDate = self._TEST_CLAIM_ADMIN_DOB
        telecom = []
        phone = ClaimAdminPractitionerConverter.build_fhir_contact_point(
            self._TEST_CLAIM_ADMIN_PHONE,
            ContactPointSystem.PHONE,
            ContactPointUse.HOME
        )
        telecom.append(phone)
        email = ClaimAdminPractitionerConverter.build_fhir_contact_point(
            self._TEST_CLAIM_ADMIN_EMAIL,
            ContactPointSystem.EMAIL,
            ContactPointUse.HOME
        )
        telecom.append(email)
        fhir_practitioner.telecom = telecom

        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/practitioner-qualification-type"
        qualification = PractitionerQualification.construct()
        qualification.code = ClaimAdminPractitionerConverter.build_codeable_concept(
            system=system,
            code="CA",
            display=_("Claim Administrator")
        )
        fhir_practitioner.qualification = [qualification]
        
        organization_reference = Reference.construct()
        resource_type = 'Organization'
        resource_id = '92345678'

        organization_reference.type = resource_type
        organization_reference.identifier = HealthFacilityOrganisationConverter.build_fhir_identifier(
            self._TEST_HF_CODE,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            R4IdentifierConfig.get_fhir_generic_type_code())
        organization_reference.reference = f'{resource_type}/{resource_id}'
        organization_reference.display = resource_id

        extension_organization = Extension.construct()
        extension_organization.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/reference"
        extension_organization.valueReference = organization_reference
        fhir_practitioner.extension = [extension_organization]

        return fhir_practitioner

    def verify_fhir_instance(self, fhir_obj):
        self.assertEqual(1, len(fhir_obj.name))
        human_name = fhir_obj.name[0]
        self.assertTrue(isinstance(human_name, HumanName))
        self.assertEqual(self.test_claim_admin.other_names, human_name.given[0])
        self.assertEqual(self.test_claim_admin.last_name, human_name.family)
        self.assertEqual("usual", human_name.use)
        for identifier in fhir_obj.identifier:
            self.assertTrue(isinstance(identifier, Identifier))
            code = ClaimAdminPractitionerConverter.get_first_coding_from_codeable_concept(identifier.type).code
            if code == R4IdentifierConfig.get_fhir_generic_type_code():
                self.assertEqual(self.test_claim_admin.code, identifier.value)
            elif code == R4IdentifierConfig.get_fhir_uuid_type_code():
                self.assertEqual(str(self.test_claim_admin.uuid), identifier.value)
        self.assertEqual(self._TEST_CLAIM_ADMIN_DOB, fhir_obj.birthDate.isoformat())
        self.assertEqual(2, len(fhir_obj.telecom))
        for telecom in fhir_obj.telecom:
            self.assertTrue(isinstance(telecom, ContactPoint))
            if telecom.system == "phone":
                self.assertEqual(self._TEST_CLAIM_ADMIN_PHONE, telecom.value)
            elif telecom.system == "email":
                self.assertEqual(self._TEST_CLAIM_ADMIN_EMAIL, telecom.value)
        self.assertEqual(1, len(fhir_obj.qualification))
        self.assertEqual("CA", fhir_obj.qualification[0].code.coding[0].code)
        self.assertEqual("Claim Administrator", fhir_obj.qualification[0].code.coding[0].display)
        self.assertEqual("92345678", fhir_obj.extension[0].valueReference.display)



from api_fhir_r4.configurations import (
    BaseApiFhirConfiguration,
    R4IdentifierConfig,
    R4LocationConfig,
    R4MaritalConfig,
    R4IssueTypeConfig,
    R4ClaimConfig,
    R4CoverageEligibilityConfiguration,
    R4CommunicationRequestConfig,
    R4OrganisationConfig,
    R4CoverageConfig,
    R4SubscriptionConfig,
    R4PaymentNoticeConfig
)


class R4ApiFhirConfig(BaseApiFhirConfiguration):

    @classmethod
    def get_identifier_configuration(cls):
        return R4IdentifierConfig

    @classmethod
    def get_location_type_configuration(cls):
        return R4LocationConfig

    @classmethod
    def get_marital_type_configuration(cls):
        return R4MaritalConfig

    @classmethod
    def get_issue_type_configuration(cls):
        return R4IssueTypeConfig

    @classmethod
    def get_claim_configuration(cls):
        return R4ClaimConfig

    @classmethod
    def get_coverage_eligibility_configuration(cls):
        return R4CoverageEligibilityConfiguration

    @classmethod
    def get_communication_request_configuration(cls):
        return R4CommunicationRequestConfig

    @classmethod
    def get_coverage_configuration(cls):
        return R4CoverageConfig

    @classmethod
    def get_organisation_configuration(cls):
        return R4OrganisationConfig

    @classmethod
    def get_subscription_configuration(cls):
        return R4SubscriptionConfig

    @classmethod
    def get_payment_notice_configuration(cls):
        return R4PaymentNoticeConfig



from django.core.exceptions import MultipleObjectsReturned
from django.http import Http404
from django.shortcuts import get_object_or_404, get_list_or_404


class DbManagerUtils(object):

    __FIRST = 0

    @classmethod
    def get_object_or_none(cls, model, **kwargs):
        try:
            result = get_object_or_404(model, **kwargs)
        except MultipleObjectsReturned:
            result = get_list_or_404(model, **kwargs)[cls.__FIRST]
        except Http404:
            result = None
        return result



import copy

from insuree.models import Family, Insuree
from api_fhir_r4.converters import GroupConverter
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.serializers import BaseFHIRSerializer
from insuree.services import FamilyService, InsureeService
from core.models import resolve_id_reference

from django.forms.models import model_to_dict

class GroupSerializer(BaseFHIRSerializer):
    fhirConverter = GroupConverter()

    def create(self, validated_data):
        #validated_data = resolve_id_reference(Family, validated_data)
        request = self.context.get("request")
        user = request.user

        insuree_id = validated_data.get('head_insuree_id')
        members_family = validated_data.pop('members_family')

        if Family.objects.filter(head_insuree_id=insuree_id).count() > 0:
            raise FHIRException('Exists family with the provided head')

        insuree = Insuree.objects.get(id=insuree_id)
        copied_data = copy.deepcopy(validated_data)
        copied_data["head_insuree"] = insuree.__dict__
        copied_data["contribution"] = None
        new_family = FamilyService(user).create_or_update(copied_data)

        # assign members of family (insuree) to the family
        for mf in members_family:
            mf = mf.__dict__
            del mf['_state']
            mf['family_id'] = new_family.id
            InsureeService(user).create_or_update(mf)

        return new_family

    def update(self, instance, validated_data):
        #validated_data = resolve_id_reference(validated_data)
        # TODO: This doesn't work
        request = self.context.get("request")
        validated_data.pop('_state')
        members_family = validated_data.pop('members_family')
        user = request.user
        head_id = validated_data.get('head_insuree_id', None)
        family_uuid = validated_data.get('uuid', None)
        if head_id:
            family = Family.objects.filter(head_insuree_id=head_id, validity_to__isnull=True).first()
            if not family:
                raise FHIRException('No family with following head id `{}`'.format(head_id))
        elif family_uuid:
            family = Family.objects.filter(uuid=family_uuid, validity_to__isnull=True).first()
            if not family:
                raise FHIRException('No family with following uuid `{}`'.format(head_id))
            
        validated_data["id"] = family.id
        validated_data["uuid"] = family.uuid
        instance = FamilyService(user).create_or_update(validated_data)
        for mf in members_family:
            mf = mf.__dict__
            mf.pop('_state')
            mf['family_id'] = instance.id
            InsureeService(user).create_or_update(mf)
            
        
        
        
        return instance



from api_fhir_r4.converters import MedicationConverter

from fhir.resources.R4B.medication import Medication
from api_fhir_r4.tests import MedicationTestMixin
from api_fhir_r4.tests.mixin import ConvertToImisTestMixin, ConvertToFhirTestMixin, ConvertJsonToFhirTestMixin


class MedicationConverterTestCase(MedicationTestMixin,
                                  ConvertToImisTestMixin,
                                  ConvertToFhirTestMixin,
                                  ConvertJsonToFhirTestMixin):
    converter = MedicationConverter
    fhir_resource = Medication
    json_repr = 'test/test_medication.json'



from django.test import TestCase

from api_fhir_r4.models import Subscription
from api_fhir_r4.services import SubscriptionService
from api_fhir_r4.utils import TimeUtils
from core.models import User


class SubscriptionServiceTest(TestCase):
    user = None
    service = None
    create_payload = {
        'status': 1,
        'channel': 0,
        'expiring': TimeUtils.str_iso_to_date('2021-01-01T01:02:03+01:00'),
        'endpoint': 'https://subscriptiontest.com/test',
        'headers': 'Authentication=Bearer 12351231',
        'criteria': {
            'resource_type': 'Patient',
            'chfid__startswith': '1',
        },
    }

    update_payload = {
        'status': 0
    }

    @classmethod
    def setUpTestData(cls):
        super(SubscriptionServiceTest, cls).setUpTestData()
        if not User.objects.filter(username='admin_sub_service').exists():
            User.objects.create_superuser(username='admin_sub_service', password='S\/pe®Pąßw0rd™')
        cls.user = User.objects.filter(username='admin_sub_service').first()
        cls.service = SubscriptionService(cls.user)

    def test_create(self):
        result = self.service.create(self.create_payload)
        self.assertIsInstance(result, dict)
        self.assertTrue(result['success'])
        self.assertEqual('Ok'.lower(), result['message'].lower())
        created_id = result['data']['id']
        self.assertTrue(created_id)
        queryset = Subscription.objects.filter(id=created_id)
        self.assertEqual(queryset.count(), 1)

    def test_update(self):
        result_create = self.service.create(self.create_payload)
        created_id = result_create['data']['id']
        update_payload = {**self.update_payload, **{'id': created_id}}
        result = self.service.update(update_payload)
        self.assertIsInstance(result, dict)
        self.assertTrue(result['success'])
        self.assertEqual('Ok'.lower(), result['message'].lower())
        changed_fields = {key: result['data'][key] for key in result['data'] if key in update_payload}
        self.assertDictEqual(update_payload, changed_fields)
        queryset = Subscription.objects.filter(id=created_id)
        self.assertEqual(queryset.count(), 1)

    def test_delete(self):
        result_create = self.service.create(self.create_payload)
        created_id = result_create['data']['id']
        delete_payload = {'id': created_id}
        result = self.service.delete(delete_payload)
        self.assertTrue(result['success'])
        self.assertEqual('Ok'.lower(), result['message'].lower())
        queryset = Subscription.objects.filter(id=created_id, is_deleted=True)
        self.assertEqual(queryset.count(), 1)
        queryset = Subscription.objects.filter(id=created_id, is_deleted=False)
        self.assertEqual(queryset.count(), 0)



from django.db import connection
from fhir.resources.R4B.coverageeligibilityresponse import (
    CoverageEligibilityResponse as FHIRCoverageEligibilityResponse,
    CoverageEligibilityResponseInsuranceItem,
    CoverageEligibilityResponseInsurance,
    CoverageEligibilityResponseInsuranceItemBenefit
)
from fhir.resources.R4B.money import Money
from fhir.resources.R4B.period import Period

from api_fhir_r4.configurations import (
    GeneralConfiguration,
    R4CoverageEligibilityConfiguration as Config
)
from api_fhir_r4.converters import (
    BaseFHIRConverter,
    PatientConverter,
    ReferenceConverterMixin
)
from api_fhir_r4.defaultConfig import DEFAULT_CFG
from api_fhir_r4.models import CoverageEligibilityRequestV2 as FHIRCoverageEligibilityRequest
from api_fhir_r4.utils import TimeUtils
from claim.models import (
    ClaimService,
    ClaimItem
)
from insuree.models import (
    Insuree,
    InsureePolicy
)
from medical.models import (
    Item,
    Service
)
from policy.models import Policy
from policy.services import (
    EligibilityRequest,
    EligibilityResponse
)
from product.models import Product, ProductService, ProductItem


class CoverageEligibilityRequestConverter(BaseFHIRConverter):

    @classmethod
    def to_fhir_obj(cls, coverage_eligibility_response, coverage_eligibility_request,
                    reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_response = cls.build_fhir_obligatory_fields(coverage_eligibility_request)
        fhir_response.patient = cls.build_fhir_patient(coverage_eligibility_request.chf_id)
        for item in coverage_eligibility_response.items:
            if item.status in Config.get_fhir_active_policy_status():
                cls.build_fhir_insurance(fhir_response, item, coverage_eligibility_request)
        return fhir_response

    @classmethod
    def to_imis_obj(cls, fhir_coverage_eligibility_request, audit_user_id):
        fhir_coverage_eligibility_request = FHIRCoverageEligibilityRequest(**fhir_coverage_eligibility_request)
        chf_id = cls.build_imis_chf(fhir_coverage_eligibility_request)
        item_code, service_code = cls.build_imis_item_service(fhir_coverage_eligibility_request)
        return EligibilityRequest(chf_id, service_code, item_code)

    @classmethod
    def build_fhir_obligatory_fields(cls, coverage_eligibility_request):
        fhir_eligibility_response = {"status": 'active', "outcome": 'complete'}

        default_insurance_organisation = DEFAULT_CFG['R4_fhir_insurance_organisation_config']
        resource_id = default_insurance_organisation['id']
        reference_insurer = {"reference": f'Organization/{resource_id}'}
        fhir_eligibility_response["insurer"] = reference_insurer

        reference_patient = {"reference": f'Patient/{coverage_eligibility_request.chf_id}'}
        fhir_eligibility_response['patient'] = reference_patient

        reference_coverage_eligibility_request = {"reference": f'CoverageEligibilityRequest'}
        fhir_eligibility_response['request'] = reference_coverage_eligibility_request

        fhir_eligibility_response["purpose"] = ["benefits"]
        fhir_eligibility_response["created"] = TimeUtils.date().isoformat()
        return FHIRCoverageEligibilityResponse(**fhir_eligibility_response)

    @classmethod
    def build_fhir_patient(cls, chf_id):
        insuree = Insuree.objects.filter(chf_id=chf_id, validity_to__isnull=True)
        if insuree.count() == 1:
            insuree = insuree.first()
            reference = PatientConverter.build_fhir_resource_reference(
                insuree,
                type='Patient',
                display=chf_id
            )
            return reference

    @classmethod
    def build_fhir_insurance(cls, fhir_response, item, request):
        result = CoverageEligibilityResponseInsurance.construct()
        cls.build_fhir_coverage(result, item.policy_uuid)
        cls.build_fhir_benefit_period(result, item.start_date, item.expiry_date)
        # get the data from SP
        try:
            with connection.cursor() as cur:
                sql = """\
                            DECLARE @MinDateService DATE, @MinDateItem DATE,
                                    @ServiceLeft INT, @ItemLeft INT,
                                    @isItemOK BIT, @isServiceOK BIT;
                            EXEC [dbo].[uspServiceItemEnquiry] @CHFID = %s, @ServiceCode = %s, @ItemCode = %s,
                                 @MinDateService = @MinDateService OUTPUT, @MinDateItem = @MinDateItem OUTPUT,
                                 @ServiceLeft = @ServiceLeft OUTPUT, @ItemLeft = @ItemLeft OUTPUT,
                                 @isItemOK = @isItemOK OUTPUT, @isServiceOK = @isServiceOK OUTPUT;
                            SELECT @MinDateService, @MinDateItem, @ServiceLeft, @ItemLeft, @isItemOK, @isServiceOK
                        """
                cur.execute(sql, (request.chf_id,
                                  request.service_code,
                                  request.item_code))
                res = cur.fetchone()  # retrieve the stored proc @Result table

                (prod_id, total_admissions_left, total_visits_left, total_consultations_left, total_surgeries_left,
                 total_deliveries_left, total_antenatal_left, consultation_amount_left, surgery_amount_left,
                 delivery_amount_left,
                 hospitalization_amount_left, antenatal_amount_left) = res
                cur.nextset()
                (min_date_service, min_date_item, service_left,
                 item_left, is_item_ok, is_service_ok) = cur.fetchone()
                response_eligibility_sp = EligibilityResponse(
                    eligibility_request=request,
                    prod_id=prod_id or None,
                    total_admissions_left=total_admissions_left or 0,
                    total_visits_left=total_visits_left or 0,
                    total_consultations_left=total_consultations_left or 0,
                    total_surgeries_left=total_surgeries_left or 0,
                    total_deliveries_left=total_deliveries_left or 0,
                    total_antenatal_left=total_antenatal_left or 0,
                    consultation_amount_left=consultation_amount_left or 0.0,
                    surgery_amount_left=surgery_amount_left or 0.0,
                    delivery_amount_left=delivery_amount_left or 0.0,
                    hospitalization_amount_left=hospitalization_amount_left or 0.0,
                    antenatal_amount_left=antenatal_amount_left or 0.0,
                    min_date_service=min_date_service,
                    min_date_item=min_date_item,
                    service_left=service_left or 0,
                    item_left=item_left or 0,
                    is_item_ok=is_item_ok is True,
                    is_service_ok=is_service_ok is True
                )
        except Exception:
            response_eligibility_sp = EligibilityResponse(
                eligibility_request=None,
                prod_id=None,
                total_admissions_left=0,
                total_visits_left=0,
                total_consultations_left=0,
                total_surgeries_left=0,
                total_deliveries_left=0,
                total_antenatal_left=0,
                consultation_amount_left=0.0,
                surgery_amount_left=0.0,
                delivery_amount_left=0.0,
                hospitalization_amount_left=0.0,
                antenatal_amount_left=0.0,
                min_date_service=None,
                min_date_item=None,
                service_left=0,
                item_left=0,
                is_item_ok=False,
                is_service_ok=False
            )

        # build coverag item - product/benefit
        result.item = []
        cls.build_fhir_benefit_item_element(result, response_eligibility_sp)
        # check services and items etc
        prod_service = ProductService.objects\
            .filter(product=prod_id, validity_to=None, service__code=request.service_code).first()
        prod_item = ProductItem.objects\
            .filter(product=prod_id, validity_to=None, item__code=request.item_code).first()
        # build coverage item - service
        if prod_service:
            cls.build_fhir_benefit_item_service_element(result, response_eligibility_sp, prod_service.service)
        # build coverage item - item
        if prod_item:
            cls.build_fhir_benefit_item_item_element(result, response_eligibility_sp, prod_item.item)
        if type(fhir_response.insurance) is not list:
            fhir_response.insurance = [result]
        else:
            fhir_response.insurance.append(result)

    @classmethod
    def build_fhir_coverage(cls, insurance, policy_uuid):
        # Due to circular dependency import has to be done inside of method
        from api_fhir_r4.converters import CoverageConverter
        policy = Policy.objects.filter(uuid__iexact=policy_uuid, validity_to__isnull=True).first()
        reference_coverage = CoverageConverter.build_fhir_resource_reference(
            policy,
            type='Coverage',
            display=policy.uuid
        )
        insurance.coverage = reference_coverage

    @classmethod
    def build_fhir_benefit_period(cls, insurance, start_date, expiry_date):
        benefit_period = Period.construct()
        benefit_period.start = start_date
        benefit_period.end = expiry_date
        insurance.benefitPeriod = benefit_period

    @classmethod
    def build_fhir_benefit_item_element(cls, insurance, response):
        item = CoverageEligibilityResponseInsuranceItem.construct()
        system = F"{GeneralConfiguration.get_system_base_url()}CodeSystem/coverage-item-category"
        item.category = cls.build_codeable_concept(
            system=system,
            code="benefit",
            display="Benefit Package"
        )
        cls.__build_item_product_name(fhir_item=item, prod_id=response.prod_id)
        item.benefit = []
        if response.total_admissions_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="admissions_left",
                display="total_admissions",
                value=response.total_admissions_left
            )
        if response.total_consultations_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="consultations_left",
                display="total_consultations",
                value=response.total_consultations_left
            )
        if response.total_surgeries_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="surgeries_left",
                display="total_surgeries",
                value=response.total_surgeries_left
            )
        if response.total_deliveries_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="deliveries_left",
                display="total_deliveries",
                value=response.total_deliveries_left
            )
        if response.total_visits_left:
            cls.build_fhir_int_item_benefit_element(
               item=item,
               code="visits_left",
               display="total_visits",
               value=response.total_visits_left
            )
        if response.total_antenatal_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="antenatal_left",
                display="total_antenatal",
                value=response.total_antenatal_left
            )
        if response.hospitalization_amount_left:
            cls.build_fhir_money_item_benefit_element(
                item=item,
                code="hospitalization_amount",
                display="hospitalization_amount",
                value=response.hospitalization_amount_left
            )
        if response.delivery_amount_left:
            cls.build_fhir_money_item_benefit_element(
                item=item,
                code="delivery_amount",
                display="delivery_amount",
                value=response.delivery_amount_left
            )
        if response.surgery_amount_left:
            cls.build_fhir_money_item_benefit_element(
                item=item,
                code="surgery_amount",
                display="surgery_amount",
                value=response.surgery_amount_left
            )
        if response.consultation_amount_left:
            cls.build_fhir_money_item_benefit_element(
                item=item,
                code="consultation_amount",
                display="consultation_amount",
                value=response.consultation_amount_left
            )
        if response.antenatal_amount_left:
            cls.build_fhir_money_item_benefit_element(
                item=item,
                code="antenatal_amount",
                display="antenatal_amount",
                value=response.antenatal_amount_left
            )
        insurance.item.append(item)

    @classmethod
    def build_fhir_benefit_item_item_element(cls, insurance, response, stored_item):
        item = CoverageEligibilityResponseInsuranceItem.construct()
        system = F"{GeneralConfiguration.get_system_base_url()}CodeSystem/coverage-item-category"
        item.category = cls.build_codeable_concept(
            system=system,
            code="item",
            display="Item"
        )
        item.benefit = []
        if response.min_date_item:
            cls.build_fhir_string_item_benefit_element(
                item=item,
                code="min_date_item",
                display="Mininum date",
                value=response.min_date_item
            )
        if response.item_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="item_left",
                display="Items left",
                value=response.service_left
            )
        code = stored_item.code
        item.productOrService = cls.build_simple_codeable_concept(code)
        item.excluded = not response.is_service_ok
        insurance.item.append(item)

    @classmethod
    def build_fhir_benefit_item_service_element(cls, insurance, response, stored_service):
        item = CoverageEligibilityResponseInsuranceItem.construct()
        system = F"{GeneralConfiguration.get_system_base_url()}CodeSystem/coverage-item-category"
        item.category = cls.build_codeable_concept(
            system=system,
            code="service",
            display="Service"
        )
        item.benefit = []
        if response.min_date_service:
            cls.build_fhir_string_item_benefit_element(
                item=item,
                code="min_date_item",
                display="Mininum date",
                value=response.min_date_item
            )
        if response.item_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="service_left",
                display="Services left",
                value=response.item_left
            )
        code = stored_service.code
        item.productOrService = cls.build_simple_codeable_concept(code)
        item.excluded = not response.is_item_ok
        insurance.item.append(item)

    @classmethod
    def build_fhir_int_item_benefit_element(cls, item, code, display, value):
        system = F"{GeneralConfiguration.get_system_base_url()}CodeSystem/coverage-item-benefit-type"
        benefit = CoverageEligibilityResponseInsuranceItemBenefit.construct()
        benefit.type = cls.build_codeable_concept(
            system=system,
            code=code,
            display=display
        )
        benefit.allowedUnsignedInt = value
        item.benefit.append(benefit)

    @classmethod
    def build_fhir_money_item_benefit_element(cls, item, code, display, value):
        system = F"{GeneralConfiguration.get_system_base_url()}CodeSystem/coverage-item-benefit-type"
        benefit = CoverageEligibilityResponseInsuranceItemBenefit.construct()
        benefit.type = cls.build_codeable_concept(
            system=system,
            code=code,
            display=display
        )
        money_value = Money.construct()
        money_value.value = value
        benefit.allowedMoney = money_value
        item.benefit.append(benefit)

    @classmethod
    def build_fhir_string_item_benefit_element(cls, item, code, display, value):
        system = F"{GeneralConfiguration.get_system_base_url()}CodeSystem/coverage-item-benefit-type"
        benefit = CoverageEligibilityResponseInsuranceItemBenefit.construct()
        benefit.type = cls.build_codeable_concept(
            system=system,
            code=code,
            display=display
        )
        benefit.allowedString = value
        item.benefit.append(benefit)

    @classmethod
    def build_imis_chf(cls, fhir_coverage_eligibility_request):
        chf_id = None
        patient_reference = fhir_coverage_eligibility_request.patient
        if patient_reference:
            chf_id = PatientConverter.get_resource_id_from_reference(patient_reference)
        return chf_id

    @classmethod
    def build_imis_item_service(cls, fhir_coverage_eligibility_request):
        service_code = None
        item_code = None
        if fhir_coverage_eligibility_request.item:
            for item in fhir_coverage_eligibility_request.item:
                type_service = cls.__get_code_from_codeable_concept_by_coding_code(item.category)
                if type_service == 'item':
                    item_code = item.productOrService.text
                if type_service == 'service':
                    service_code = item.productOrService.text
        return item_code, service_code

    @classmethod
    def __get_code_from_codeable_concept_by_coding_code(cls, codeable_concept):
        service_code = None
        if codeable_concept:
            coding = cls.get_first_coding_from_codeable_concept(codeable_concept)
            if coding:
                service_code = coding.code
        return service_code

    @classmethod
    def __get_coverage_data(cls, response):
        policy = InsureePolicy.objects.filter(
            insuree__chf_id=response.eligibility_request.chf_id,
            validity_to__isnull=True).first().policy
        product = Product.objects.get(id=response.prod_id, validity_to__isnull=True)
        return policy, product

    @classmethod
    def __build_item_product_name(cls, fhir_item, prod_id):
        product_queryset = Product.objects.all().filter(id=prod_id, validity_to=None)
        product = product_queryset.first()
        fhir_item.name = product.name
        fhir_item.description = product.code



import logging
from typing import Union

from django.http.response import HttpResponseBase
from fhir.resources.R4B import FHIRAbstractModel
from rest_framework import serializers

from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.converters import BaseFHIRConverter, OperationOutcomeConverter, ReferenceConverterMixin
from core.models import User, TechnicalUser



logger = logging.getLogger(__name__)


class BaseFHIRSerializer(serializers.Serializer):
    fhirConverter = BaseFHIRConverter()

    def __init__(self, *args, **kwargs):
        self._reference_type = kwargs.pop('reference_type', ReferenceConverterMixin.UUID_REFERENCE_TYPE)
        super().__init__(*args, **kwargs)

    def to_representation(self, obj):
        try:
            if isinstance(obj, HttpResponseBase):
                return OperationOutcomeConverter.to_fhir_obj(obj).dict()
            elif isinstance(obj, FHIRAbstractModel):
                return obj.dict()
            return self.fhirConverter.to_fhir_obj(obj, self.reference_type).dict()
        except Exception as e:
            from django.conf import settings
            if settings.DEBUG:
                self._print_debug_log(e)
            raise e

    def to_internal_value(self, data):
        audit_user_id = self.get_audit_user_id()
        return self.fhirConverter.to_imis_obj(data, audit_user_id).__dict__

    def create(self, validated_data):
        raise NotImplementedError('`create()` must be implemented.')  # pragma: no cover

    def update(self, instance, validated_data):
        raise NotImplementedError('`update()` must be implemented.')  # pragma: no cover

    def get_audit_user_id(self):
        request = self.context.get("request")
        # Taking the audit_user_id from the query doesn't seem wise but there might be a use for it
        # audit_user_id = request.query_params.get('auditUserId', None)
        audit_user_id = request.user.id_for_audit if request.user else None
        if audit_user_id is None:
            audit_user_id = GeneralConfiguration.get_default_audit_user_id()
        if isinstance(audit_user_id, int):
            return audit_user_id
        else:
            return self.__get_technical_audit_user(audit_user_id)

    @property
    def reference_type(self):
        return self._reference_type

    @reference_type.getter
    def reference_type(self):
        return self._reference_type

    @reference_type.setter
    def reference_type(self, reference_type: Union[ReferenceConverterMixin.UUID_REFERENCE_TYPE,
                                                   ReferenceConverterMixin.CODE_REFERENCE_TYPE,
                                                   ReferenceConverterMixin.DB_ID_REFERENCE_TYPE]):
        self._reference_type = reference_type

    def __get_technical_audit_user(self, technical_user_uuid):
        technical_user = TechnicalUser.objects.get(id=technical_user_uuid)
        core_user = User.objects.get(t_user=technical_user_uuid)
        interactive_user = core_user.i_user
        return interactive_user.id if interactive_user else technical_user.id_for_audit

    def _print_debug_log(self, e):
        import traceback
        debug_log = "FHIR Mapping for Serializer {self} has failed with exception {e}. Traceback: \n" \
                    + str(traceback.format_stack())
        logger.debug(debug_log)


from api_fhir_r4.serializers.patientSerializer import PatientSerializer
from api_fhir_r4.serializers.groupSerializer import GroupSerializer
from api_fhir_r4.serializers.policyHolderOrganisationSerializer import PolicyHolderOrganisationSerializer
from api_fhir_r4.serializers.contractSerializer import ContractSerializer
from api_fhir_r4.serializers.locationSerializer import LocationSerializer
from api_fhir_r4.serializers.locationSiteSerializer import LocationSiteSerializer
from api_fhir_r4.serializers.claimAdminPractitionerRoleSerializer import ClaimAdminPractitionerRoleSerializer
from api_fhir_r4.serializers.claimAdminPractitionerSerializer import ClaimAdminPractitionerSerializer
from api_fhir_r4.serializers.coverageEligibilityRequestSerializer import CoverageEligibilityRequestSerializer
# from api_fhir_r4.serializers.policyCoverageEligibilityRequestSerializer import PolicyCoverageEligibilityRequestSerializer
from api_fhir_r4.serializers.claimResponseSerializer import ClaimResponseSerializer
from api_fhir_r4.serializers.communicationRequestSerializer import CommunicationRequestSerializer
from api_fhir_r4.serializers.medicationSerializer import MedicationSerializer
from api_fhir_r4.serializers.activityDefinitionSerializer import ActivityDefinitionSerializer
from api_fhir_r4.serializers.insurancePlanSerializer import InsurancePlanSerializer
from api_fhir_r4.serializers.codeSystemSerializer import CodeSystemSerializer
from api_fhir_r4.serializers.healthFacilityOrganisationSerializer import HealthFacilityOrganisationSerializer
from api_fhir_r4.serializers.enrolmentOfficerPractitionerSerializer import EnrolmentOfficerPractitionerSerializer
from api_fhir_r4.serializers.enrolmentOfficerPractitionerRoleSerializer import \
    EnrolmentOfficerPractitionerRoleSerializer
from api_fhir_r4.serializers.communicationSerializer import CommunicationSerializer
from api_fhir_r4.serializers.invoiceSerializer import InvoiceSerializer
from api_fhir_r4.serializers.insuranceOrganisationSerializer import InsuranceOrganizationSerializer
from api_fhir_r4.serializers.billSerializer import BillSerializer
from api_fhir_r4.serializers.claimSerializer import ClaimSerializer



from api_fhir_r4.converters import EnrolmentOfficerPractitionerRoleConverter
from fhir.resources.R4B.practitionerrole import PractitionerRole
from api_fhir_r4.tests import EnrolmentOfficerPractitionerRoleTestMixin
from api_fhir_r4.tests.mixin import ConvertToImisTestMixin, ConvertToFhirTestMixin, ConvertJsonToFhirTestMixin


class EnrolmentOfficerPractitionerRoleConverterTestCase(EnrolmentOfficerPractitionerRoleTestMixin,
                                                        ConvertToImisTestMixin,
                                                        ConvertToFhirTestMixin,
                                                        ConvertJsonToFhirTestMixin):
    converter = EnrolmentOfficerPractitionerRoleConverter
    fhir_resource = PractitionerRole
    json_repr = 'test/test_enrolmentOfficerPractitionerRole.json'
    def setUp(self):
        super(EnrolmentOfficerPractitionerRoleConverterTestCase, self).setUp()
        self.sub_str[self._TEST_OFFICER_UUID] = self.test_officer.uuid



import logging
from abc import ABC, abstractmethod
from typing import List, Callable, Iterable, Union
from django.db import models

from fhir.resources.R4B.resource import Resource

from api_fhir_r4.converters import BaseFHIRConverter, ReferenceConverterMixin
from api_fhir_r4.exceptions import FHIRException
from fhir.resources.R4B import FHIRAbstractModel


DEFAULT_REF_TYPE = ReferenceConverterMixin.UUID_REFERENCE_TYPE
logger = logging.getLogger(__name__)


class _ConverterWrapper:

    def __init__(self, converter: BaseFHIRConverter):
        self.converter = converter

    def to_imis(self, resource: Union[Iterable[Resource], Resource], reference_type, audit_user):
        return self.__convert_to_imis(self.converter.to_imis_obj, resource, reference_type, [audit_user])

    def to_fhir(self, resource, reference_type):
        return self.__convert_to_fhir(self.converter.to_fhir_obj, resource, [reference_type])

    def __convert_to_fhir(self, method, resource, args):
        if not resource:
            return []

        try:
            if isinstance(resource, Iterable) and   'resourceType' not in resource:
                return [method(next_resource, *args) for next_resource in resource]
            else:
                return [method(resource, *args)]
        except BaseException as e:
            logger.error(f"Failed to process resource ({resource.__class__}/{resource.id if hasattr(resource, 'id') else '' }, "
                         f"reason: {e}")
            self.__raise_default_exception(resource, e)

    def __convert_to_imis(self, method, resource, reference_type, args):
        try:
            if isinstance(resource, Iterable) and 'resourceType' not in resource :
                return [
                    self.__convert_single_resource(next_, method, args, reference_type) for next_ in resource
                ]
            else:
                return [self.__convert_single_resource(resource, method, args, reference_type)]
        except BaseException as e:
            logger.error(f"Failed to process resource ({resource.get('resourceType')}/{resource.get('id')}, "
                         f"reason: {e}")
            self.__raise_default_exception(resource, e)

    def __convert_single_resource(self, resource, method, args, ref_type):
        converted = method(resource, *args)
        self.__bind_uuid_to_converted_resource(converted, resource, ref_type)
        return converted

    def __bind_uuid_to_converted_resource(self, converted, contained_fhir_resource: dict, reference_type):
        """
        By default, converters doesn't bind uuid to created resource. In that if id is explicitly given in contained
        resource this information will be lost in the process. Identifiers assigned from contained resource definitions
        are necessary for using contained resource in process of creating object from them. This is only available for
        UUID type identifiers.
        """
        assert reference_type == DEFAULT_REF_TYPE, \
            f'Invalid reference type, assigning contained resource uuid explicitly is available only for ' \
            f'{DEFAULT_REF_TYPE}'

        assert contained_fhir_resource.get('id') is not None, \
            F'Resources created from contained data requires non empty ID field.'
        converted.uuid = contained_fhir_resource['id']

    def __raise_default_exception(self, resource, error):
        raise FHIRException(
            "Failed to process resource: {}, using converter {}. Error: {}".format(
                resource, self.converter, str(error)
            )
        )


class FHIRContainedResourceConverter:
    def __init__(self, imis_resource_name, converter, resource_extract_method=None, reference_type=DEFAULT_REF_TYPE):
        """
        Parameters
        ----------
        :param imis_resource_name: Name of attribute, which value should be transformed.
        :param converter: FHIR Converter, it's used for mapping imis resource to fhir representation.
        It must implement at least to_fhir_obj() method.
        :param resource_extract_method: Optional argument. Function used for getting attribute value from IMIS Model.
        It has two arguments, first is django model, second one is imis_resource_name. Default function is
        imis_model.__getattribute__(imis_resource_name). Return type can be model or iterable (e.g. list of attributes).
        :param reference_type: Optional argument. Determine what object value will be used as reference and id.
        """
        self.imis_resource_name = imis_resource_name
        self.extract_value = resource_extract_method or (lambda model, attribute: model.__getattribute__(attribute))
        self.converter = _ConverterWrapper(converter)
        self.reference_type = reference_type

    def convert(self, imis_obj: models.Model) -> List[FHIRAbstractModel]:
        """Convert IMIS Model attribute to FHIR Object.

        :param imis_obj: IMIS Object with attribute that have to be converted.
        :return: Attribute converted to FHIR object list. If attribute is single object then it's still converted to
        list format.
        """
        resource = self.extract_value(imis_obj, self.imis_resource_name)
        return self.converter.to_fhir(resource, self.reference_type)


class IMISContainedResourceConverter:
    def __init__(self, resource_reference_type, converter, reference_type=DEFAULT_REF_TYPE):
        """
        Parameters
        ----------
        :param resource_reference_type: Contained resource type
        :param converter: FHIR Converter, it's used for mapping fhir resource to imis representation.
        It must implement at least to_imis_obj() method.
        :param resource_extract_method: Optional argument. Function used for getting attribute value from IMIS Model.
        It has two arguments, first is django model, second one is imis_resource_name. Default function is
        imis_model.__getattribute__(imis_resource_name). Return type can be model or iterable (e.g. list of attributes).
        :param reference_type: Optional argument. Determine what object value will be used as reference and id.
        """
        self.fhir_resource_type = resource_reference_type
        self.converter = _ConverterWrapper(converter)
        self.reference_type = reference_type

    def convert(self, fhir_dict_repr: dict, audit_user_id: int = None) -> List[models.Model]:
        """Extracts FHIR contained resource based on resource_type and converts to IMIS object.

        :param fhir_dict_repr: FHIR Dict representation with contained key that have to be converted.
        :param audit_user_id: Id of user performing given operation.
        :return: Attribute converted to FHIR object list. If attribute is single object then it's still converted to
        list format.
        """
        resource = None
        if fhir_dict_repr:
            if 'resourceType' in fhir_dict_repr and fhir_dict_repr['resourceType'] == self.fhir_resource_type:
                resource = fhir_dict_repr
            else:
                resource = [r for r in fhir_dict_repr.get('contained', {}) if r['resourceType'] == self.fhir_resource_type]
            if resource:
                return self.converter.to_imis(resource, self.reference_type, audit_user_id)
        
        return None


from django.utils.translation import gettext as _
from django.utils.translation import gettext
from location.models import Location

from api_fhir_r4.configurations import R4IdentifierConfig, R4LocationConfig
from api_fhir_r4.converters import BaseFHIRConverter, ReferenceConverterMixin
from fhir.resources.R4B.location import Location as FHIRLocation

from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.mapping.locationMapping import LocationTypeMapping
from api_fhir_r4.utils import DbManagerUtils


class LocationConverter(BaseFHIRConverter, ReferenceConverterMixin):
    PHYSICAL_TYPES = LocationTypeMapping.PHYSICAL_TYPES_DEFINITIONS

    @classmethod
    def to_fhir_obj(cls, imis_location, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_location = FHIRLocation.construct()
        cls.build_fhir_physical_type(fhir_location, imis_location)
        cls.build_fhir_pk(fhir_location, imis_location, reference_type)
        cls.build_fhir_location_identifier(fhir_location, imis_location)
        cls.build_fhir_location_name(fhir_location, imis_location)
        cls.build_fhir_part_of(fhir_location, imis_location, reference_type)
        cls.build_fhir_status(fhir_location, imis_location)
        cls.build_fhir_mode(fhir_location)
        return fhir_location

    @classmethod
    def to_imis_obj(cls, fhir_location, audit_user_id):
        errors = []
        fhir_location = FHIRLocation(**fhir_location)
        imis_location = Location()
        cls.build_imis_location_identiftier(imis_location, fhir_location, errors)
        cls.build_imis_location_name(imis_location, fhir_location, errors)
        cls.build_imis_location_type(imis_location, fhir_location, errors)
        cls.build_imis_parent_location_id(imis_location, fhir_location, errors)
        cls.check_errors(errors)
        return imis_location

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_location_code_type()

    @classmethod
    def get_reference_obj_uuid(cls, imis_location: Location):
        return imis_location.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_location: Location):
        return imis_location.id

    @classmethod
    def get_reference_obj_code(cls, imis_location: Location):
        return imis_location.code

    @classmethod
    def get_fhir_resource_type(cls):
        return Location

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        return DbManagerUtils.get_object_or_none(Location, **cls.get_database_query_id_parameteres_from_reference(reference))

    @classmethod
    def build_fhir_location_identifier(cls, fhir_location, imis_location):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_location)
        fhir_location.identifier = identifiers
        cls._validate_imis_identifiers(identifiers)

    @classmethod
    def build_fhir_location_code_identifier(cls, identifiers, imis_location):
        if imis_location is not None:
            identifier = cls.build_fhir_identifier(imis_location.code,
                                                   R4IdentifierConfig.get_fhir_identifier_type_system(),
                                                   R4IdentifierConfig.get_fhir_location_code_type())
            identifiers.append(identifier)

    @classmethod
    def build_imis_location_identiftier(cls, imis_location, fhir_location, errors):
        value = cls.get_fhir_identifier_by_code(
            fhir_location.identifier,
            R4IdentifierConfig.get_fhir_location_code_type()
        )
        if value:
            imis_location.code = value
        cls.valid_condition(imis_location.code is None, gettext('Missing location code'), errors)

    @classmethod
    def build_fhir_location_name(cls, fhir_location, imis_location):
        fhir_location.name = imis_location.name
        cls._validate_imis_name(imis_location)

    @classmethod
    def build_imis_location_name(cls, imis_location, fhir_location, errors):
        name = fhir_location.name
        if not cls.valid_condition(name is None, _('Missing location `name` attribute'), errors):
            imis_location.name = name

    @classmethod
    def build_fhir_physical_type(cls, fhir_location, imis_location):
        cls._validate_physical_type(imis_location)
        system_definition = cls.PHYSICAL_TYPES.get(imis_location.type)
        fhir_location.physicalType = cls.build_codeable_concept(**system_definition)

    @classmethod
    def build_imis_location_type(cls, imis_location, fhir_location, errors):
        code = None
        try:
            code = fhir_location.physicalType.coding[0].code
            imis_code = next(
                (k for k, v in cls.PHYSICAL_TYPES.items() if v['code'] == code), None
            )
            imis_location.type = imis_code
            cls._validate_physical_type(imis_location)

        except KeyError:
            errors.append(_('Missing location physical type'))
        except FHIRException:
            errors.append(
                _('Invalid location physical type with code %(code), allowed codes are: %(types)') % {
                    'code': code,
                    'types': [v['code'] for v in cls.PHYSICAL_TYPES.values]
                }
            )

    @classmethod
    def build_fhir_part_of(cls, fhir_location, imis_location, reference_type):
        if not cls.__is_highers_level_location(imis_location):
            fhir_location.partOf = LocationConverter.build_fhir_resource_reference(
                imis_location.parent,
                'Location',
                imis_location.parent.code,
                reference_type=reference_type
            )

    def get_location_from_address(cls, fhir_patient_address):
        location_reference_ext = next((
            ext for ext in fhir_patient_address.extension if 'address-location-reference' in ext.url
        ))
        location =  cls.get_imis_obj_by_fhir_reference(location_reference_ext.valueReference)
        if location is None and location_reference_ext:
            raise FHIRException(f"Invalid location reference, {location_reference_ext.valueReference} doesn't match any location.")
        if location is None:
            matching_locations = Location.objects \
                .filter(
                    validity_to__isnull=True,
                    name=address.district,
                    parent__name=address.state,
                    type="D"  # HF is expected to be at district level
                ).distinct()\
                .all()
        
            if matching_locations.count() != 1:
                raise FHIRException(cls.__get_invalid_location_msg(address, matching_locations))
            else:
                location = matching_locations.first()
        return location
    
    
    @classmethod
    def __get_invalid_location_msg(cls, address, matching_locations):
        count = matching_locations.count()
        if count == 0:
            return _(F"No matching location for district {address.district}, state {address.state}.")
        elif count > 1:
            return _(F"More than one matching location district {address.district}, state {address.state}:\n"
                     F"{matching_locations}.")

    @classmethod
    def build_imis_parent_location_id(cls, imis_location, fhir_location, errors):
        if fhir_location.partOf:
            resource = cls.get_imis_obj_by_fhir_reference(fhir_location.partOf) 
            if not resource:
                errors.append(
                    _('Invalid Location\'s partOf.reference')
                )


    @classmethod
    def _validate_imis_identifiers(cls, identifiers):
        code_identifier = cls.get_fhir_identifier_by_code(identifiers, cls.get_fhir_code_identifier_type())
        if not code_identifier:
            raise FHIRException(_('Code identifier has to be present in FHIR Location identifiers, '
                                  '\nidentifiers are: %(identifiers)') % {'identifiers': identifiers})

    @classmethod
    def _validate_imis_name(cls, imis_location):
        if not imis_location.name:
            raise FHIRException(
                _('Location  %(location_uuid)s without name') % {'uuid': imis_location.uuid}
            )

    @classmethod
    def _validate_physical_type(cls, imis_location):
        allowed_keys = cls.PHYSICAL_TYPES.keys()
        if not imis_location or (imis_location.type not in allowed_keys):
            
            error_msg_keys = {
                'type': imis_location.type, 'location': imis_location.uuid, 'types': allowed_keys
            } if imis_location else {'type':None,'location':None,'types': allowed_keys}
            raise FHIRException(
                _('Invalid Location\'s type %(type) for location %(location), '
                  'supported types are: %(types)') % error_msg_keys
            )

    @classmethod
    def build_fhir_mode(cls, fhir_location):
        fhir_location.mode = 'instance'

    @classmethod
    def build_fhir_status(cls, fhir_location, imis_location):
        if cls.__location_active(imis_location):
            fhir_location.status = R4LocationConfig.get_fhir_code_for_active()
        else:
            fhir_location.status = R4LocationConfig.get_fhir_code_for_inactive()

    @classmethod
    def __location_active(cls, imis_location):
        return imis_location.validity_to is None

    @classmethod
    def __is_highers_level_location(cls, imis_location):
        return imis_location.parent is None

    @classmethod
    def _is_code_reference(cls, location_part_of):
        if location_part_of.identifier:
            try:
                ref_type_code = location_part_of.identifier.type.coding[0].code
                return ref_type_code == R4IdentifierConfig.get_fhir_location_code_type()
            except KeyError:
                raise FHIRException(_("Invalid format for Location's partOf identifier."))
        else:
            # Default expected reference identifier is uuid
            return False



from django.contrib.contenttypes.models import ContentType
from fhir.resources.R4B.paymentnotice import PaymentNotice

from .errors import (
    ERROR_BAD_STATUS,
    ERROR_BAD_PAYMENT_STATUS,
    ERROR_BAD_TYPE_REQUEST
)
from api_fhir_r4.converters import (
    BaseFHIRConverter,
    ReferenceConverterMixin
)
from api_fhir_r4.paymentNotice.mapping import (
    PaymentNoticeStatusMapping,
    PaymentNoticePaymentDetailStatusMapping,
    PaymentNoticePaymentStatusMapping
)
from invoice.models import (
    PaymentInvoice,
    DetailPaymentInvoice
)

PAYMENT_STATUS = ['cleared', 'paid']
STATUS = ['active', 'cancelled', 'draft', 'entered-in-error']
SUBJECT_TYPE = ['bill', 'invoice']


class PaymentNoticeToImisConverter(BaseFHIRConverter, ReferenceConverterMixin):
    @classmethod
    def to_imis_obj(cls, fhir_payment_notice, audit_user_id=None):
        errors = []
        fhir_payment_notice = PaymentNotice(**fhir_payment_notice)
        imis_payment = PaymentInvoice()
        imis_payment_detail = DetailPaymentInvoice()
        cls.build_imis_payment_date_created(imis_payment, imis_payment_detail, fhir_payment_notice)
        cls.build_imis_payment_payer_ref(imis_payment,fhir_payment_notice)
        cls.build_imis_payment_reconciliation_status(imis_payment, fhir_payment_notice, errors)
        cls.build_imis_payment_amount(imis_payment, imis_payment_detail, fhir_payment_notice)
        cls.build_imis_payment_date_payment(imis_payment, fhir_payment_notice)
        cls.build_imis_payment_json_ext(imis_payment, fhir_payment_notice)
        cls.build_imis_payment_detail_invoice(imis_payment_detail, fhir_payment_notice, errors)
        cls.build_imis_payment_detail_status(imis_payment_detail, fhir_payment_notice, errors)
        invoice_status = None
        if imis_payment_detail.subject_id:
            invoice_status = cls.get_imis_invoice_status(fhir_payment_notice, errors)
        cls.check_errors(errors)
        imis_payment.imis_payment_detail = imis_payment_detail
        imis_payment.invoice_status = invoice_status
        return imis_payment
    @classmethod
    def build_imis_payment_payer_ref(cls, imis_payment, fhir_payment_notice):
        imis_payment.payer_ref = fhir_payment_notice.payment.reference

    @classmethod
    def build_imis_payment_date_created(cls, imis_payment, imis_payment_detail, fhir_payment_notice):
        created = fhir_payment_notice.created
        imis_payment.date_created = created
        imis_payment.date_updated = created
        imis_payment_detail.date_created = created
        imis_payment_detail.date_updated = created

    @classmethod
    def build_imis_payment_reconciliation_status(cls, imis_payment, fhir_payment_notice, errors):
        coding = cls.get_first_coding_from_codeable_concept(fhir_payment_notice.paymentStatus)
        payment_status = coding.code
        if payment_status in PAYMENT_STATUS:
            imis_payment.reconciliation_status = PaymentNoticePaymentStatusMapping\
                .to_imis_status[payment_status]
        else:
            errors.append(ERROR_BAD_PAYMENT_STATUS)

    @classmethod
    def get_imis_invoice_status(cls, fhir_payment_notice, errors):
        status = fhir_payment_notice.status
        if status in STATUS:
            invoice_status = PaymentNoticeStatusMapping \
                .to_imis_status[status]
            return invoice_status
        else:
            errors.append(ERROR_BAD_STATUS)
            return None

    @classmethod
    def build_imis_payment_detail_status(cls, imis_payment_detail, fhir_payment_notice, errors):
        if fhir_payment_notice.status in STATUS:
            imis_payment_detail.status = PaymentNoticePaymentDetailStatusMapping.\
                to_imis_status[fhir_payment_notice.status]
        else:
            errors.append(ERROR_BAD_STATUS)

    @classmethod
    def build_imis_payment_json_ext(cls, imis_payment, fhir_payment_notice):
        # for saving PaymentReconciliation in json_ext
        reconciliation_id = cls.get_id_from_reference(fhir_payment_notice.payment)
        imis_payment.json_ext = {'reconciliation': {'id': reconciliation_id}}

    @classmethod
    def build_imis_payment_detail_invoice(cls, imis_payment_detail, fhir_payment_notice, errors):
        subject_id = cls.get_id_from_reference(fhir_payment_notice.request)
        subject_type = fhir_payment_notice.request.type.lower()
        if subject_type in SUBJECT_TYPE:
            imis_payment_detail.subject_type = cls._convert_content_type(subject_type)
            imis_payment_detail.subject_id = subject_id
        else:
            errors.append(ERROR_BAD_TYPE_REQUEST)
        

    @classmethod
    def _convert_content_type(cls, subject_type):
        return ContentType.objects.get(model__iexact=subject_type)

    @classmethod
    def build_imis_payment_amount(cls, imis_payment, imis_payment_detail, fhir_payment_notice):
        amount = fhir_payment_notice.amount.value
        imis_payment.amount_received = amount
        imis_payment_detail.amount = amount

    @classmethod
    def build_imis_payment_date_payment(cls, imis_payment, fhir_payment_notice):
        payment_date = fhir_payment_notice.paymentDate
        imis_payment.date_payment = payment_date



from api_fhir_r4.serializers import CodeSystemSerializer

from rest_framework import viewsets
from rest_framework.response import Response
from rest_framework.permissions import IsAuthenticated
from api_fhir_r4.permissions import FHIRApiInsureePermissions

from rest_framework.views import APIView
from api_fhir_r4.views import CsrfExemptSessionAuthentication
from django.core.exceptions import PermissionDenied


class CodeSystemOpenIMISPatientRelationshipViewSet(viewsets.ViewSet):

    serializer_class = CodeSystemSerializer
    permission_classes = (IsAuthenticated,)
    authentication_classes = [CsrfExemptSessionAuthentication] + APIView.settings.DEFAULT_AUTHENTICATION_CLASSES

    def list(self, request):
        # we don't use typical instance, we only indicate the model and the field to be mapped into CodeSystem
        if not request.user.has_perms(FHIRApiInsureePermissions.permissions_get):
            raise PermissionDenied("unauthorized")
        serializer = CodeSystemSerializer(
            instance=None,
            **{
                "model_name": 'Relation',
                "code_field": 'id',
                "display_field": 'relation',
                "id": 'patient-contact-relationship',
                "name": 'PatientContactRelationshipCS',
                "title": 'Contact Relationship (Patient)',
                "description": "Indicates the Relationship of a Patient with the Head of the Family. "
                               "Values defined by openIMIS.",
                "url": self.request.build_absolute_uri()
            }
        )
        data = serializer.to_representation(obj=None)
        return Response(data)



from enum import Enum
from functools import lru_cache


class ImisMaritalStatus(Enum):
    MARRIED = "M"
    SINGLE = "S"
    DIVORCED = "D"
    WIDOWED = "W"
    NOT_SPECIFIED = "N"


class ImisHfLevel(Enum):
    HEALTH_CENTER = "C"
    HOSPITAL = "H"
    DISPENSARY = "D"


class ImisLocationType(Enum):
    REGION = "R"
    DISTRICT = "D"
    WARD = "W"
    VILLAGE = "V"


class ImisClaimIcdTypes(Enum):
    ICD_0 = "icd_0"
    ICD_1 = "icd_1"
    ICD_2 = "icd_2"
    ICD_3 = "icd_3"
    ICD_4 = "icd_4"


class ImisCategoryDefinition(Enum):
    CATEGORY_SURGERY = "Surgery"
    CATEGORY_DELIVERY = "Delivery"
    CATEGORY_ANTENATAL = "Antenatal"
    CATEGORY_HOSPITALIZATION = "Hospitalization"
    CATEGORY_CONSULTATION = "Consultation"
    CATEGORY_OTHER = "Other"
    CATEGORY_VISIT = "Visit"

    @classmethod
    def get_category_display(cls, category_char):
        @lru_cache(maxsize=1)
        def cached_categories():
            return {
                "S": cls.CATEGORY_SURGERY.value,
                "D": cls.CATEGORY_DELIVERY.value,
                "A": cls.CATEGORY_ANTENATAL.value,
                "H": cls.CATEGORY_HOSPITALIZATION.value,
                "C": cls.CATEGORY_CONSULTATION.value,
                "O": cls.CATEGORY_OTHER.value,
                "V": cls.CATEGORY_VISIT.value,
            }
        try:
            cached_categories()[category_char]
        except KeyError as e:
            raise ValueError(
                f"Invalid category code: {category_char}, available categories are: \n{cached_categories()}"
            )


class BundleType(Enum):
    """
    fhir.resources.R4B doesn't use enum for bundle type,
    see https://github.com/nazrulworld/fhir.resources.R4B/blob/91bf2064aa03c6f3c252d26d7945fa0bb140b03a/fhir/resources/bundle.py#L103
    this class allows to use constants instead of plain strings.
    """
    DOCUMENT = "document"
    MESSAGE = "message"
    TRANSACTION = "transaction"
    TRANSACTION_RESPONSE = "transaction-response"
    BATCH = "batch"
    BATCH_RESPONSE = "batch-response"
    HISTORY = "history"
    SEARCHSET = "searchset"
    COLLECTION = "collection"


class ContactPointSystem(Enum):
    PHONE = "phone"
    FAX = "fax"
    EMAIL = "email"
    PAGER = "pager"
    URL = "url"
    SMS = "sms"
    OTHER = "other"


class ContactPointUse(Enum):
    HOME = "home"
    WORK = "work"
    TEMP = "temp"
    OLD = "old"
    MOBILE = "mobile"


class AddressType(Enum):
    POSTAL = "postal"
    PHYSICAL = "physical"
    BOTH = "both"



from api_fhir_r4.configurations import R4ClaimConfig, GeneralConfiguration
from claim.models import Claim
from django.utils.translation import gettext as _
from urllib.parse import urljoin


class ClaimPriorityMapping:
    SYSTEM = 'http://terminology.hl7.org/CodeSystem/processpriority'

    fhir_priority_coding = {
        'normal': {
            'system': SYSTEM,
            'code': 'normal',
            'display': 'Normal',
        }
    }


class ClaimVisitTypeMapping:
    SYSTEM = urljoin(GeneralConfiguration.get_system_base_url(), R4ClaimConfig.get_fhir_claim_visit_type_system())

    fhir_claim_visit_type_coding = {
        'E': {
            'system': SYSTEM,
            'code': 'E',
            'display': 'Emergency',
        },
        'R': {
            'system': SYSTEM,
            'code': 'R',
            'display': 'Referrals',
        },
        'O': {
            'system': SYSTEM,
            'code': 'O',
            'display': 'Other',
        }
    }


class ClaimItemCategoryMapping:
    SYSTEM = urljoin(GeneralConfiguration.get_system_base_url(), R4ClaimConfig.get_fhir_claim_item_category_system())

    fhir_claim_item_type_coding = {
        R4ClaimConfig.get_fhir_claim_item_code(): {
            'system': SYSTEM,
            'code': R4ClaimConfig.get_fhir_claim_item_code(),
            'display': 'Item',
        },
        R4ClaimConfig.get_fhir_claim_service_code(): {
            'system': SYSTEM,
            'code': R4ClaimConfig.get_fhir_claim_service_code(),
            'display': 'Service',
        }
    }


class ClaimResponseMapping(object):

    claim_status_system = f'{GeneralConfiguration.get_system_base_url()}CodeSystem/claim-status'
    claim_status = {
        f"{Claim.STATUS_REJECTED}": R4ClaimConfig.get_fhir_claim_status_rejected_code(),
        f"{Claim.STATUS_ENTERED}": R4ClaimConfig.get_fhir_claim_status_entered_code(),
        f"{Claim.STATUS_CHECKED}": R4ClaimConfig.get_fhir_claim_status_checked_code(),
        f"{Claim.STATUS_PROCESSED}": R4ClaimConfig.get_fhir_claim_status_processed_code(),
        f"{Claim.STATUS_VALUATED}": R4ClaimConfig.get_fhir_claim_status_valuated_code()
    }

    claim_outcome = {
        f"{Claim.STATUS_REJECTED}": _("complete"),
        f"{Claim.STATUS_ENTERED}": _("queued"),
        f"{Claim.STATUS_CHECKED}": _("partial"),
        f"{Claim.STATUS_PROCESSED}": _("partial"),
        f"{Claim.STATUS_VALUATED}": _("complete")
    }

    visit_type_system = f'{GeneralConfiguration.get_system_base_url()}CodeSystem/claim-visit-type'
    visit_type = {
        "E": _("Emergency"),
        "R": _("Referrals"),
        "O": _("Other"),
    }

    rejection_reason_system = f'{GeneralConfiguration.get_system_base_url()}CodeSystem/claim-rejection-reasons'
    rejection_reason = {
        -2: _("REJECTED BY AI EVALUATION"),
        -1: _("REJECTED BY MEDICAL OFFICER"),
        0: _("ACCEPTED"),
        1: _("INVALID ITEM OR SERVICE"),
        2: _("NOT IN PRICE LIST"),
        3: _("NO PRODUCT FOUND"),
        4: _("CATEGORY LIMITATION"),
        5: _("FREQUENCY FAILURE"),
        6: _("DUPLICATED"),
        7: _("FAMILY"),
        8: _("DIAGNOSIS NOT IN LIST"),
        9: _("TARGET DATE"),
        10: _("CARE TYPE"),
        11: _("MAX HOSPITAL ADMISSIONS"),
        12: _("MAX VISITS"),
        13: _("MAX CONSULTATIONS"),
        14: _("MAX SURGERIES"),
        15: _("MAX DELIVERIES"),
        16: _("QTY OVER LIMIT"),
        17: _("WAITING PERIOD FAIL"),
        19: _("MAX ANTENATAL"),
    }




from api_fhir_r4.configurations import R4PaymentNoticeConfig
from invoice.models import DetailPaymentInvoice


class PaymentNoticePaymentDetailStatusMapping:
    _IMIS_DETAIL_STATUS_ACCEPTED = DetailPaymentInvoice.DetailPaymentStatus.ACCEPTED
    _FHIR_DETAIL_STATUS_ACTIVE = R4PaymentNoticeConfig.get_fhir_payment_notice_status_active()

    _IMIS_DETAIL_STATUS_CANCELLED = DetailPaymentInvoice.DetailPaymentStatus.CANCELLED
    _FHIR_DETAIL_STATUS_CANCELLED = R4PaymentNoticeConfig.get_fhir_payment_notice_status_cancelled()

    _IMIS_DETAIL_STATUS_REJECTED = DetailPaymentInvoice.DetailPaymentStatus.REJECTED
    _FHIR_DETAIL_STATUS_ENTERED_IN_ERROR = R4PaymentNoticeConfig.get_fhir_payment_notice_status_entered_in_error()

    to_fhir_status = {
        _IMIS_DETAIL_STATUS_ACCEPTED: _FHIR_DETAIL_STATUS_ACTIVE,
        _IMIS_DETAIL_STATUS_CANCELLED: _FHIR_DETAIL_STATUS_CANCELLED,
        _IMIS_DETAIL_STATUS_REJECTED: _FHIR_DETAIL_STATUS_ENTERED_IN_ERROR,
    }

    to_imis_status = {
        _FHIR_DETAIL_STATUS_ACTIVE: _IMIS_DETAIL_STATUS_ACCEPTED,
        _FHIR_DETAIL_STATUS_CANCELLED: _IMIS_DETAIL_STATUS_CANCELLED,
        _FHIR_DETAIL_STATUS_ENTERED_IN_ERROR: _IMIS_DETAIL_STATUS_REJECTED,
    }



class FhirUtils(object):
    __ARRAY_ID_OFFSET = 1  # used to start iterating from 1

    @classmethod
    def get_next_array_sequential_id(cls, array):
        if array is None:
            array = []
        return len(array) + cls.__ARRAY_ID_OFFSET

    @classmethod
    def get_attr(cls, obj, field):
        if isinstance(obj, dict):
            return obj[field]
        else:
            return getattr(obj, field)



from api_fhir_r4.configurations import BaseConfiguration, GeneralConfiguration, R4ApiFhirConfig


class ModuleConfiguration(BaseConfiguration):
    @classmethod
    def build_configuration(cls, cfg):
        GeneralConfiguration.build_configuration(cfg)
        cls.get_r4().build_configuration(cfg)

    @classmethod
    def get_r4(cls):
        return R4ApiFhirConfig



from rest_framework.exceptions import APIException

from invoice.models import (
    PaymentInvoice,
    DetailPaymentInvoice
)
from invoice.services import PaymentInvoiceService


class PaymentNoticeSerializerUtils(object):
    _ERROR_WHILE_SAVING = 'Error while saving a payment notice: %(msg)s'
    _PAYMENT_CANCEL_STATUS = [
        DetailPaymentInvoice.DetailPaymentStatus.CANCELLED,
        DetailPaymentInvoice.DetailPaymentStatus.REJECTED
    ]

    @classmethod
    def get_result_object(cls, result):
        if result.get('success', False):
            return PaymentInvoice.objects.get(id=result['data']['id'])
        else:
            raise APIException(cls._ERROR_WHILE_SAVING % {'msg': result.get('message', 'Unknown')})

    @classmethod
    def change_invoice_status(
        cls, payment_invoice_service: PaymentInvoiceService, payment: PaymentInvoice
    ):
        payment_detail = DetailPaymentInvoice.objects.filter(payment=payment).first()
        status = payment_detail.status
        if status == DetailPaymentInvoice.DetailPaymentStatus.ACCEPTED:
            payment_invoice_service.payment_received(payment, DetailPaymentInvoice.DetailPaymentStatus.ACCEPTED)
        elif status in cls._PAYMENT_CANCEL_STATUS:
            payment_invoice_service.payment_cancelled(payment)
